import logging
import re

from warn.cache import Cache as BaseCache

from .urls import urls

logger = logging.getLogger(__name__)


class Cache(BaseCache):
    """A custom cache for Job Center sites."""

    def save(self, url, params, html):
        """Save file to the cache."""
        cache_key = self.key_from_url(url, params)
        self.write(cache_key, html)
        logger.debug(f"Saved to cache: {cache_key}")

    def fetch(self, url, params):
        """Fetch file from the cache."""
        cache_key = self.key_from_url(url, params)
        content = self.read(cache_key)
        logger.debug(f"Fetched from cache: {cache_key}")
        return content

    def key_from_url(self, url, params=None):
        """Convert a URL to a cache key."""
        page_type = (
            "records" if re.search(r"warn_lookups/\d+$", url) else "search_results"
        )
        if page_type == "records":
            record_number = url.rsplit("/")[-1]
            cache_key = f"records/{record_number}.html"
        # Otherwise this is an initial search with params or a downstream page URL
        else:
            start_key = "q[notice_on_gteq]"
            end_key = "q[notice_on_lteq]"
            # For downstream pages, extract start/end dates + page number
            if "page" in url:
                parsed_params = urls.parse_url_query(url)
                page_num = urls.page_num_from_url(url)
                start = parsed_params[start_key][0]
                end = parsed_params[end_key][0]
            # For initial search page, get metadata from params
            else:
                if not params:
                    params = {}
                start = params[start_key]
                end = params[end_key]
                page_num = 1
            cache_key = f"search_results/{start}_{end}_page{page_num}.html"
        return cache_key



import csv
import logging
import os
import typing
from pathlib import Path

import requests
from openpyxl import load_workbook

logger = logging.getLogger(__name__)


# The default home directory, if nothing is provided by the user
WARN_USER_DIR = Path(os.path.expanduser("~"))
WARN_DEFAULT_OUTPUT_DIR = WARN_USER_DIR / ".warn-scraper"

# Set the home directory
WARN_OUTPUT_DIR = Path(os.environ.get("WARN_OUTPUT_DIR", WARN_DEFAULT_OUTPUT_DIR))

# Set the subdirectories for other bits
WARN_CACHE_DIR = WARN_OUTPUT_DIR / "cache"
WARN_DATA_DIR = WARN_OUTPUT_DIR / "exports"
WARN_LOG_DIR = WARN_OUTPUT_DIR / "logs"


def create_directory(path: Path, is_file: bool = False):
    """Create the filesystem directories for the provided Path objects.

    Args:
        path (Path): The file path to create directories for.
        is_file (bool): Whether or not the path leads to a file (default: False)
    """
    # Get the directory path
    if is_file:
        # If it's a file, take the parent
        directory = path.parent
    else:
        # Other, assume it's a directory and we're good
        directory = path

    # If the path already exists, we're good
    if directory.exists():
        return

    # If not, lets make it
    logger.debug(f"Creating directory at {directory}")
    directory.mkdir(parents=True)


def write_rows_to_csv(output_path: Path, rows: list, mode="w"):
    """Write the provided list to the provided path as comma-separated values.

    Args:
        rows (list): the list to be saved
        output_path (Path): the Path were the result will be saved
        mode (str): the mode to be used when opening the file (default 'w')
    """
    create_directory(output_path, is_file=True)
    logger.debug(f"Writing {len(rows)} rows to {output_path}")
    with open(output_path, mode, newline="") as f:
        writer = csv.writer(f)
        writer.writerows(rows)


def write_dict_rows_to_csv(output_path, headers, rows, mode="w", extrasaction="raise"):
    """Write the provided dictionary to the provided path as comma-separated values.

    Args:
        output_path (Path): the Path were the result will be saved
        headers (list): a list of the headers for the output file
        rows (list): the dict to be saved
        mode (str): the mode to be used when opening the file (default 'w')
        extrasaction (str): what to do if the if a field isn't in the headers (default 'raise')
    """
    create_directory(output_path, is_file=True)
    logger.debug(f"Writing {len(rows)} rows to {output_path}")
    with open(output_path, mode, newline="") as f:
        # Create the writer object
        writer = csv.DictWriter(f, fieldnames=headers, extrasaction=extrasaction)
        # If we are writing a new row ...
        if mode == "w":
            # ... drop in the headers
            writer.writeheader()
        # Loop through the dicts and write them in one by one.
        for row in rows:
            writer.writerow(row)


def get_all_scrapers():
    """Get all the states and territories that have scrapers.

    Returns: List of lower-case post abbreviations.
    """
    this_dir = Path(__file__).parent
    scrapers_dir = this_dir / "scrapers"
    return sorted(
        p.stem for p in scrapers_dir.glob("*.py") if "__init__.py" not in str(p)
    )


def get_url(
    url, user_agent="Big Local News (biglocalnews.org)", session=None, **kwargs
):
    """Request the provided URL and return a response object.

    Args:
        url (str): the url to be requested
        user_agent (str): the user-agent header passed with the request (default: biglocalnews.org)
        session: a session object to use when making the request. optional
    """
    logger.debug(f"Requesting {url}")

    if "headers" not in kwargs:
        kwargs["headers"] = {}
    kwargs["headers"]["User-Agent"] = user_agent

    if session is not None:
        logger.debug(f"Requesting with session {session}")
        response = session.get(url, **kwargs)
    else:
        response = requests.get(url, **kwargs)
    logger.debug(f"Response code: {response.status_code}")
    return response


def parse_excel(excel_path: Path, keep_header: bool = True) -> typing.List[typing.List]:
    """Parse the Excel file at the provided path.

    Args:
        excel_path (Path): The path to an XLSX file
        keep_header (bool): Whether or not to return the header row. Default  True.

    Returns: List of values ready to write.
    """
    # Open it up
    workbook = load_workbook(filename=excel_path)

    # Get the first sheet
    worksheet = workbook.worksheets[0]

    # Convert the sheet to a list of lists
    row_list = []
    for i, r in enumerate(worksheet.rows):
        # Skip the header row, if that's what the user wants
        if i == 0 and not keep_header:
            continue

        # Parse cells
        cell_list = [cell.value for cell in r]

        # Skip empty rows
        try:
            # A list with only empty cells will throw an error
            next(c for c in cell_list if c)
        except StopIteration:
            continue

        # Add to the master list
        row_list.append(cell_list)

    # Pass it back
    return row_list



import logging
import re
from pathlib import Path

import pdfplumber
from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "chriszs"]
__tags__ = ["html", "pdf"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Ohio.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Open the cache
    cache = Cache(cache_dir)

    # Get the HTML
    base_url = "https://jfs.ohio.gov/warn/"
    r = utils.get_url(f"{base_url}index.stm")
    html = r.text

    # Save it to the cache
    state_code = "oh"
    cache.write(f"{state_code}/index.html", html)

    # Get the list of links
    soup = BeautifulSoup(html, "html5lib")
    link_list = soup.find(class_="warnYears").find_all("a")
    href_lookup = {a.text: a["href"] for a in link_list}

    # Loop through years and add any missing to the lookup
    most_recent_year = int(list(href_lookup.keys())[0])
    earliest_year = 2015  # We expect files to be available for at least 2015
    for year in range(earliest_year, most_recent_year):
        if str(year) not in href_lookup:
            href_lookup[str(year)] = f"{base_url}WARN{year}.stm"

    row_list = []

    # Loop through the links and scrape the data
    for year, href in href_lookup.items():
        # Form the URL
        if href.startswith("http"):
            url = href.replace("http://", "https://")
        else:
            url = f"{base_url}{href.replace('./', '')}"

        # If URL does not contain current or archive, assume it's a PDF
        if "archive" not in url and "current" not in url:
            cache_key = f"{state_code}/{year}.pdf"
        else:
            cache_key = f"{state_code}/{year}.html"

        # Get the file
        file_path = cache.download(cache_key, url)

        if str(file_path).endswith(".pdf"):
            # Parse the PDF
            row_list += _parse_pdf(file_path)
        else:
            # Parse the HTML
            row_list += _parse_html(file_path)

    header = list(filter(_is_header, row_list))[0]
    output_rows = [header] + list(filter(lambda row: not _is_header(row), row_list))

    # Write out
    data_path = data_dir / f"{state_code}.csv"
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the CSV
    return data_path


def _parse_html(file_path: Path) -> list:
    """
    Parse the table from the HTML.

    Keyword arguments:
    file_path -- the Path to the HTML file

    Returns: a list of lists of strings
    """
    # Get the HTML
    html = open(file_path).read()

    # Parse table
    soup = BeautifulSoup(html, "html.parser")
    table_list = soup.find_all("table")

    # We expect the first table to be there with our data
    assert len(table_list) > 0
    table = table_list[1]

    # Parse the cells
    row_list = []
    for row in table.find_all("tr"):
        cell_list = row.find_all(["th", "td"])
        if not cell_list:
            continue
        cell_list = [c.text.strip() for c in cell_list]
        row_list.append(cell_list)

    # Return it
    return row_list


def _parse_pdf(pdf_path: Path) -> list:
    """
    Parse a PDF file and return a list of rows.

    Keyword arguments:
    pdf_path -- the Path to the PDF file

    Returns: a list of rows
    """
    # Loop through the PDF pages and scrape out the data
    row_list: list = []

    with pdfplumber.open(pdf_path) as pdf:
        for page in pdf.pages:
            rows = page.extract_table({"explicit_horizontal_lines": page.curves})

            for row_index, row in enumerate(rows):
                output_row = []
                for col_index, column in enumerate(row):
                    clean_text = _clean_text(column)

                    # If cell is empty, copy from the cell above it
                    # to deal with merged cells. Except for numbers, e.g. of employees,
                    # which we don't want to double count.
                    if (
                        clean_text == ""
                        and row_index > 0
                        and col_index < len(row_list[row_index - 1])
                        and row_list[row_index - 1][col_index] is not None
                        and not row_list[row_index - 1][col_index].strip().isnumeric()
                    ):
                        clean_text = row_list[row_index - 1][col_index]

                    output_row.append(clean_text)

                row_list.append(output_row)

    return row_list


def _clean_text(text: str) -> str:
    """
    Clean up text from a PDF cell.

    Keyword arguments:
    text -- the text to clean

    Returns: the cleaned text
    """
    if text is None:
        return ""
    # Collapse newlines
    partial = re.sub(r"\n", " ", text)
    # Standardize whitespace
    return re.sub(r"\s+", " ", partial)


def _is_header(row: list) -> bool:
    """
    Determine if a row is a header.

    Keyword arguments:
    row -- the row to check

    Returns: True if the row is a header, False otherwise
    """
    return (
        row[0].startswith("Date Rec")
        or row[0].startswith("DateReceived")
        or row[0].endswith("WARN Notices")
    )


if __name__ == "__main__":
    scrape()



import logging
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Indiana.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Open the cache
    cache = Cache(cache_dir)

    # Get the HTML
    latest_url = "https://www.in.gov/dwd/2567.htm"
    r = utils.get_url(latest_url)
    latest_html = r.text
    cache.write("in/latest.html", latest_html)

    # Parse tables
    latest_soup = BeautifulSoup(latest_html, "html.parser")
    latest_tables = latest_soup.find_all("table")

    # Scrape table
    output_rows = []
    for i, table in enumerate(latest_tables):
        row_list = _parse_table(table, include_headers=i == 0)
        logger.debug(f"Scraped {len(row_list)} rows latest table {i+1}")
        output_rows.extend(row_list)

    # Get the archive tables
    archive_url = "https://www.in.gov/dwd/3125.htm"
    r = utils.get_url(archive_url)
    archive_html = r.text
    cache.write("in/archive.html", archive_html)

    # Parse tables
    archive_soup = BeautifulSoup(archive_html, "html.parser")
    archive_tables = archive_soup.find_all("table")

    # Scrape table
    for i, table in enumerate(archive_tables):
        row_list = _parse_table(table, include_headers=False)
        logger.debug(f"Scraped {len(row_list)} rows latest table {i+1}")
        output_rows.extend(row_list)

    # Write out
    data_path = data_dir / "in.csv"
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the CSV
    return data_path


def _parse_table(table, include_headers) -> list:
    # Parse the cells
    row_list = []
    tags = [
        "td",
    ]
    if include_headers:
        tags.append("th")
    for row in table.find_all("tr"):
        cell_list = row.find_all(tags)
        if not cell_list:
            continue
        cell_list = [c.text.strip() for c in cell_list]
        row_list.append(cell_list)

    # Return it
    return row_list


if __name__ == "__main__":
    scrape()



import logging
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "shallotly"]
__tags__ = ["html", "csv"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Virginia.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Get the WARN page
    url = "https://www.vec.virginia.gov/warn-notices"
    r = utils.get_url(url)
    html = r.text

    # Save it to the cache
    cache = Cache(cache_dir)
    cache.write("va/source.html", html)

    # Parse out the CSV download link
    soup = BeautifulSoup(html, "html.parser")
    csv_href = soup.find("a", text="Download")["href"]
    csv_url = f"https://www.vec.virginia.gov{csv_href}"

    # Download it to the cache
    cache.download("va/source.csv", csv_url)

    # Open it up as a list of rows
    csv_rows = cache.read_csv("va/source.csv")

    # Set the export path
    data_path = data_dir / "va.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, csv_rows)

    # Return the export path
    return data_path


if __name__ == "__main__":
    scrape()



import logging
from datetime import datetime
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = [
    "zstumgoren",
    "Dilcia19",
    "stucka",
]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Connecticut.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Open the cache
    cache = Cache(cache_dir)

    # We start in 2015
    current_year = datetime.now().year

    # Get the full range of years
    year_range = range(2015, current_year + 1)

    output_rows = []
    for year in year_range:
        url = f"https://www.ctdol.state.ct.us/progsupt/bussrvce/warnreports/warn{year}.htm"
        cache_key = f"ct/{year}.html"

        if cache.exists(cache_key) and year < current_year:
            html = cache.read(cache_key)
        else:
            r = utils.get_url(url)
            html = r.text
            cache.write(cache_key, html)

        # Parse out the table
        soup = BeautifulSoup(html, "html.parser")
        if year == 2016:
            table = soup.find_all("table", "style15")
        else:
            table = soup.find_all("table", "MsoNormalTable")

        # Parse out the data
        row_list = _scrape_table(table)

        # Add data to the big list
        output_rows.extend(row_list)

    # Tack headers on the top
    header_row = [
        "warn_date",
        "affected_company",
        "layoff_location",
        "number_workers",
        "layoff_date",
        "closing",
        "closing_date",
        "union",
        "union_address",
    ]
    row_list = [header_row] + output_rows

    # Set the export path
    data_path = data_dir / "ct.csv"

    # Write out to csv
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path
    return data_path


def _scrape_table(table) -> list:
    """Scrape the provided table.

    Returns: List of data rows.
    """
    row_list = []
    # loop over table to process each row, skipping the header
    for table_row in table[0].find_all("tr")[1:]:

        # Get all the cells
        table_cells = table_row.find_all("td")

        # if a row has more than 9 cells it is handled separately
        # the 2016 table has some cells with nested tags
        if len(table_cells) > 9:
            output_row = _problem_cells(table_cells)
            row_list.append(output_row)
            continue
        # if a row has less than 9 it is skipped because it is incomplete
        elif len(table_cells) < 9:
            continue

        # for the rest, loop over cells for each row
        output_row = []
        for table_cell in table_cells:
            cell = table_cell.text.strip()
            cell = " ".join(cell.split())
            output_row.append(cell)

        # test to see if the row is blank
        if not output_row:
            continue

        # Add row to the big list
        row_list.append(output_row)

    # Pass it back
    logger.debug(f"{len(row_list)} rows parsed")
    return row_list


def _problem_cells(table_cells):
    """Deal with problem rows in the 2016 table."""
    output_row = []
    for table_cell in table_cells:
        current_cell = table_cell.text.strip()
        current_cell = " ".join(current_cell.split())
        if table_cells.index(table_cell) == 0:
            output_row.append(current_cell)
        else:
            previous_index = table_cells.index(table_cell) - 1
            previous_cell = table_cells[previous_index].text.strip()
            previous_cell = " ".join(previous_cell.split())
            if current_cell == previous_cell:
                continue
            else:
                output_row.append(current_cell)
    return output_row


if __name__ == "__main__":
    scrape()



from pathlib import Path

import pytest

from warn.platforms.job_center.cache import Cache

from .conftest import write_file


@pytest.fixture
def cache_dir(tmp_path):
    """Create a path for a cache test."""
    return str(tmp_path.joinpath("ks"))


@pytest.mark.parametrize(
    "test_input",
    [
        # Initial search uses kwargs on base search url
        (
            "https://www.kansasworks.com/search/warn_lookups",
            {"q[notice_on_gteq]": "2020-05-01", "q[notice_on_lteq]": "2021-07-02"},
            "search_results/2020-05-01_2021-07-02_page1.html",
        ),
        # downstream search result page
        (
            "https://www.kansasworks.com/search/warn_lookups?commit=Search&page=2&q%5Bemployer_name_cont%5D=&q%5Bmain_contact_contact_info_addresses_full_location_city_matches%5D=&q%5Bnotice_eq%5D=true&q%5Bnotice_on_gteq%5D=2020-05-01&q%5Bnotice_on_lteq%5D=2021-07-02&q%5Bservice_delivery_area_id_eq%5D=&q%5Bzipcode_code_start%5D=&utf8=%E2%9C%93",
            {},
            "search_results/2020-05-01_2021-07-02_page2.html",
        ),
        # record page
        ("https://www.kansasworks.com/search/warn_lookups/123", {}, "records/123.html"),
    ],
)
def test_cache_key_from_url(test_input, cache_dir):
    """Test if the cache key is what's expected."""
    url, params, expected_key = test_input
    cache = Cache(cache_dir)
    cache_key = cache.key_from_url(url, params=params)
    assert cache_key == expected_key


@pytest.mark.parametrize(
    "test_input",
    [
        # downstream search result page
        (
            "https://www.kansasworks.com/search/warn_lookups?commit=Search&page=2&q%5Bemployer_name_cont%5D=&q%5Bmain_contact_contact_info_addresses_full_location_city_matches%5D=&q%5Bnotice_eq%5D=true&q%5Bnotice_on_gteq%5D=2020-05-01&q%5Bnotice_on_lteq%5D=2021-07-02&q%5Bservice_delivery_area_id_eq%5D=&q%5Bzipcode_code_start%5D=&utf8=%E2%9C%93",
            {},
            "search_results/2020-05-01_2021-07-02_page2.html",
        ),
        # record page
        ("https://www.kansasworks.com/search/warn_lookups/123", {}, "records/123.html"),
    ],
)
def test_cache_save(test_input, cache_dir):
    """Test saving data to the cache."""
    url, params, expected_name = test_input
    cache = Cache(cache_dir)
    cache.save(url, params, "<html><h1>Hello world</h1><html>")
    files = [str(f) for f in Path(cache_dir).glob("**/*.html")]
    assert len(files) == 1
    assert files[0].endswith(expected_name)


@pytest.mark.parametrize(
    "test_input",
    [
        # downstream search result page
        (
            "https://www.kansasworks.com/search/warn_lookups?commit=Search&page=2&q%5Bemployer_name_cont%5D=&q%5Bmain_contact_contact_info_addresses_full_location_city_matches%5D=&q%5Bnotice_eq%5D=true&q%5Bnotice_on_gteq%5D=2020-05-01&q%5Bnotice_on_lteq%5D=2021-07-02&q%5Bservice_delivery_area_id_eq%5D=&q%5Bzipcode_code_start%5D=&utf8=%E2%9C%93",
            {},
            "search_results/2020-05-01_2021-07-02_page2.html",
        ),
        # record page
        ("https://www.kansasworks.com/search/warn_lookups/123", {}, "records/123.html"),
    ],
)
def test_cache_fetch(test_input, cache_dir):
    """Test retrieving data from the cache."""
    url, params, name = test_input
    expected_content = "<html><h1>Hello world</h1><html>"
    dest = Path(cache_dir, name)
    write_file(dest, expected_content)
    cache = Cache(cache_dir)
    content = cache.fetch(url, params)
    assert content == expected_content



import logging
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Utah.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Open the cache
    cache = Cache(cache_dir)

    # Get the HTML
    url = "https://jobs.utah.gov/employer/business/warnnotices.html"
    r = utils.get_url(url)
    html = r.text
    cache.write("ut/source.html", html)

    # Parse table
    soup = BeautifulSoup(html, "html.parser")
    table_list = soup.find_all("table")

    # Scrape out the data
    row_list = []
    for i, table in enumerate(table_list):
        row_list.extend(_parse_table(table, i == 0))

    # Write out
    data_path = data_dir / "ut.csv"
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the CSV
    return data_path


def _parse_table(table, include_headers) -> list:
    # Parse the cells
    row_list = []
    tags = ["td"]
    if include_headers:
        tags.append("th")
    for row in table.find_all("tr"):
        cell_list = row.find_all(tags)
        if not cell_list:
            continue
        cell_list = [c.text.strip() for c in cell_list]
        row_list.append(cell_list)

    # Return it
    return row_list


if __name__ == "__main__":
    scrape()



import urllib.parse


class urls:
    """Create a URL object."""

    @classmethod
    def parse_url_query(cls, url):
        """Parse the querystring from a URL."""
        return urllib.parse.parse_qs(url)

    @classmethod
    def page_num_from_url(cls, url):
        """Pull page number from URL."""
        queries = cls.parse_url_query(url)
        return int(queries["page"][0])  # it's a one-element array



import shutil
from pathlib import Path

import pytest

# NOTE: To check if vcrpy/pytest-vcr
# is using cassettes as opposed to making
# live web requests, uncomment below
# and pass pytest caplog fixture to
# a test function. More details here:
#    https://vcrpy.readthedocs.io/en/latest/debugging.html
# import vcr
# import logging
# Initialize logging in order see output from vcrpy
# logging.basicConfig()
# vcr_log = logging.getLogger("vcr")
# vcr_log.setLevel(logging.INFO)


@pytest.fixture
def warn_scraper_dir(tmp_path):
    """Set scraper directory."""
    return str(tmp_path.joinpath("warn-scraper"))


@pytest.fixture
def cache_dir(warn_scraper_dir):
    """Set scraper directory."""
    return str(Path(warn_scraper_dir).joinpath("cache"))


@pytest.fixture
def create_scraper_dir(warn_scraper_dir):
    """Create scraper directory."""
    Path(warn_scraper_dir).mkdir(parents=True, exist_ok=True)


@pytest.fixture
def create_cache_dir(cache_dir):
    """Create cache directory."""
    Path(cache_dir).mkdir(parents=True, exist_ok=True)


@pytest.fixture
def copy_html_to_cache(cache_dir):
    """Copy HTML content to cache directory."""
    fl_dir = Path(cache_dir).joinpath("fl")
    fl_dir.mkdir(parents=True, exist_ok=True)
    for html_fixture in ["2021_page_1.html", "2021_page_2.html"]:
        src = Path(__file__).parent.joinpath("fixtures").joinpath(html_fixture)
        dest = fl_dir.joinpath(html_fixture)
        shutil.copy(src, dest)


@pytest.fixture
def set_etl_dir_env_var(warn_scraper_dir, monkeypatch):
    """Set the environment variable for the output directory."""
    monkeypatch.setenv("WARN_ETL_DIR", warn_scraper_dir)


def read_fixture(file_name):
    """Read in provided fixture."""
    path = str(Path(__file__).parent.joinpath("fixtures").joinpath(file_name))
    return file_contents(path)


def file_contents(pth):
    """Read provided file path."""
    with open(pth, newline="") as f:
        return f.read()


def write_file(pth, contents):
    """Write to provided file path."""
    Path(pth).parent.mkdir(parents=True, exist_ok=True)
    with open(pth, "w", newline="") as f:
        f.write(contents)



import logging
import re
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils

__authors__ = ["zstumgoren", "Dilcia19"]
__tags__ = [
    "html",
]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Alabama.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    output_csv = data_dir / "al.csv"
    page = utils.get_url("https://www.madeinalabama.com/warn-list/")
    # can't see 2020 listings when I open web page, but they are on the summary in the google search
    soup = BeautifulSoup(page.text, "html.parser")
    table = soup.find_all("table")  # output is list-type
    table_rows = table[0].find_all("tr")
    # Handle the header
    raw_header = table_rows.pop(0)
    header_row = _extract_fields_from_row(raw_header, "th")
    output_rows = [header_row]
    # Process remaining rows
    discarded_rows = []
    for table_row in table_rows:
        # Discard bogus data lines (see last lines of source data)
        # based on check of first field ("Closing or Layoff")
        data = _extract_fields_from_row(table_row, "td")
        layoff_type = data[0]
        if re.match(r"(clos|lay)", layoff_type, re.I):
            output_rows.append(data)
        else:
            discarded_rows.append(data)
    if discarded_rows:
        logger.warn(f"Warning: Discarded {len(discarded_rows)} dirty data row(s)")
    utils.write_rows_to_csv(output_csv, output_rows)
    return output_csv


def _extract_fields_from_row(row, element):
    """Pluck data from the provided row and element."""
    row_data = []
    fields = row.find_all(element)
    for raw_field in fields:
        field = raw_field.text.strip()
        row_data.append(field)
    return row_data


if __name__ == "__main__":
    scrape()



import re
from datetime import datetime
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["chriszs"]
__tags__ = ["html"]


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Georgia.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    state_code = "ga"
    cache = Cache(cache_dir)

    # The basic configuration for the scrape
    base_url = "https://www.dol.state.ga.us/public/es/warn/searchwarns/list"

    area = 9  # statewide

    current_year = datetime.now().year
    first_year = 2002  # first available year

    years = list(range(first_year, current_year + 1))
    years.reverse()

    # Loop through the years and scrape them one by one
    output_rows = []
    for i, year in enumerate(years):
        # Concoct the URL
        url = f"{base_url}?geoArea={area}&year={year}&step=search"
        cache_key = f"{state_code}/{year}.html"

        # Read from cache if available and not this year or the year before
        if cache.exists(cache_key) and year < current_year - 1:
            html = cache.read(cache_key)
        else:
            # Otherwise, go request it
            page = utils.get_url(url)
            html = page.text
            cache.write(cache_key, html)

        # Scrape out the table
        new_rows = _parse_table(
            html,
            "emplrList",
            include_headers=i == 0,  # After the first loop, we can skip the headers
        )

        # Concatenate the rows
        output_rows.extend(new_rows)

    # Write out the results
    data_path = data_dir / f"{state_code}.csv"
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the CSV
    return data_path


def _parse_table(html, id, include_headers=True):
    """
    Parse HTML table with given ID.

    Keyword arguments:
    html -- the HTML to parse
    id -- the ID of the table to parse
    include_headers -- whether to include the headers in the output (default True)

    Returns: a list of rows
    """
    # Parse out data table
    soup = BeautifulSoup(html, "html.parser")
    table_list = soup.find_all(id=id)  # output is list-type

    # We expect the first table to be there with our data
    assert len(table_list) > 0
    table = table_list[0]

    output_rows = []
    column_tags = ["td"]

    if include_headers:
        column_tags.append("th")

    # Loop through the table and grab the data
    for table_row in table.find_all("tr"):
        columns = table_row.find_all(column_tags)
        output_row = []

        for column in columns:
            # Collapse newlines
            partial = re.sub(r"\n", " ", column.text)
            # Standardize whitespace
            clean_text = re.sub(r"\s+", " ", partial).strip()
            output_row.append(clean_text)

        # Skip any empty rows
        if len(output_row) == 0 or output_row == [""]:
            continue

        output_rows.append(output_row)

    return output_rows


if __name__ == "__main__":
    scrape()



import logging
import os
import re
from datetime import datetime
from pathlib import Path

import pdfplumber
from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["chriszs"]
__tags__ = ["html", "pdf"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Lousiana.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Fire up the cache
    cache = Cache(cache_dir)

    # The basic configuration for the scrape
    state_code = "la"
    base_url = "https://www.laworks.net/"
    file_base = "Downloads_WFD"

    # Download the root page
    url = f"{base_url}Downloads/{file_base}.asp"
    html = utils.get_url(url).text

    # Save it to the cache
    cache_key = f"{state_code}/{file_base}.html"
    cache.write(cache_key, html)

    # Parse out the links to WARN notice PDFs
    document = BeautifulSoup(html, "html.parser")
    links = document.find_all("a")

    all_rows = []

    for link in links:
        if "WARN Notices" in link.text:
            # Download the PDF
            pdf_url = f"{base_url}{link['href']}"
            pdf_path = _read_or_download(cache, state_code, pdf_url)

            # Process the PDF
            rows = _process_pdf(pdf_path)
            all_rows.extend(rows)

    # Insert a header row with clean column names.
    # We are here assuming that the columns don't change between years
    # and that one that contains "Employees Affected" will be clean.
    output_rows = [list(filter(_is_clean_header, all_rows))[0]] + list(
        filter(lambda row: not _is_header(row), all_rows)
    )

    # Write out to CSV
    data_path = data_dir / f"{state_code}.csv"
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path
    return data_path


def _append_contents_to_cells_in_row_above(rows: list, index: int, row: list) -> list:
    """
    Append the contents of a row to the contents of the cell above it.

    Keyword arguments:
    index -- the index of the row
    row -- the row to check
    rows -- the rows to append to
    """
    for column_index, cell in enumerate(row):
        if _cell_above_exists(column_index, rows):
            rows[len(rows) - 1][column_index].extend(cell)
    return rows


def _cell_above_exists(column_index: int, rows: list) -> bool:
    """
    Return True if the cell above the current cell exists.

    Keyword arguments:
    column_index -- the index of the column
    rows -- the rows to check
    """
    return column_index < len(rows[len(rows) - 1])


def _has_rows(rows: list) -> bool:
    """
    Determine if the table has rows.

    Keyword arguments:
    rows -- the rows to check

    Returns: True if the table has rows, False otherwise
    """
    return len(rows) > 0


def _is_first(index: int) -> bool:
    """
    Determine if a row is the first row in a table.

    Keyword arguments:
    index -- the index of the row

    Returns: True if the row is the first row in a table, False otherwise
    """
    return index == 0


def _is_mostly_empty(row: list) -> bool:
    """
    Check if a row has few populated cells. Used to determine if carried over from a previous page.

    Keyword arguments:
    row -- the row to check

    Returns: True if the row is mostly empty, False otherwise
    """
    return len(list(filter(pdfplumber.utils.extract_text, row))) <= 2


def _process_pdf(pdf_path: Path) -> list:
    """
    Process a PDF file.

    Keyword arguments:
    pdf_path -- the path to the PDF file

    Returns: a list of rows
    """
    output_rows: list = []

    with pdfplumber.open(pdf_path) as pdf:
        for page in pdf.pages:
            for table in page.debug_tablefinder().tables:
                for index, row in enumerate(table.rows):
                    row = [_extract_cell_chars(page, cell) for cell in row.cells]

                    # If the first row in a table is mostly empty,
                    # append its contents to the previous row
                    if (
                        _is_first(index)
                        and _is_mostly_empty(row)
                        and _has_rows(output_rows)
                    ):
                        output_rows = _append_contents_to_cells_in_row_above(
                            output_rows, index, row
                        )
                    # Otherwise, append the row
                    else:
                        output_rows.append(row)

    return _clean_rows(output_rows)


def _clean_rows(rows: list) -> list:
    """
    Clean up rows.

    Keyword arguments:
    rows -- the rows to clean

    Returns: the cleaned rows
    """
    output_rows = []

    for row in rows:
        output_row = []
        for column_index, chars in enumerate(row):
            text = _clean_text(pdfplumber.utils.extract_text(chars))

            # If we're on the first column, try to extract location and notes
            if _is_first(column_index):
                # Tries to extract a company name, appends it to the row
                company_name = _extract_company_name(chars)
                output_row.append(company_name)
                remaining_text = text.replace(company_name, "")

                # Tries to extract a note, typically UPDATE or WARN RESCINDED
                note = _extract_note(chars).strip()

                # Whatever is left is assumbed to be the location
                location = remaining_text.strip().replace(note, "")

                # Append the location and note to the row or headers for those
                if _is_header(output_row):
                    output_row.append("Location")
                    output_row.append("Note")
                else:
                    output_row.append(location)
                    output_row.append(note)
            else:
                # Appends the remaining text to the row
                output_row.append(text)

        output_rows.append(output_row)

    return output_rows


def _extract_cell_chars(page: pdfplumber.pdf.Page, bbox: tuple) -> list:
    """
    Extract the characters from a cell.

    Keyword arguments:
    page -- the page from which to extract the characters
    bbox -- the bounding box of the cell

    Returns: a list of characters
    """
    # If the bounding box is empty, append an empty list
    if bbox is None:
        return []

    # Expand the bounding box to ensure it encompasses the bottom line of text
    vertical_threshold = 5
    expanded_bbox = _vertically_expand_bounding_box(bbox, vertical_threshold)

    # Get the characters from the cell
    return page.within_bbox(expanded_bbox).chars


def _vertically_expand_bounding_box(bbox: tuple, increase: int) -> tuple:
    """
    Expand the bounding box by a given amount in the vertical direction.

    Keyword arguments:
    bbox -- the bounding box to expand
    increase -- the amount to expand the bounding box

    Returns: the expanded bounding box
    """
    return (
        bbox[0],
        bbox[1],
        bbox[2],
        bbox[3] + increase,
    )


def _read_or_download(cache: Cache, prefix: str, url: str) -> Path:
    """
    Read a file from the cache or downloads it.

    Keyword arguments:
    cache -- the cache to use
    prefix -- the prefix to use for the cache key
    url -- the URL to download

    Returns: the path to the file
    """
    file_name = os.path.basename(url)
    cache_key = f"{prefix}/{file_name}"

    exists = cache.exists(cache_key)
    year = _extract_year(file_name)
    current_year = datetime.now().year

    # Form a file path so we can read from the cache
    if exists and year < current_year - 1:
        return cache.path / cache_key

    return cache.download(cache_key, url)


def _extract_year(text: str) -> int:
    """
    Extract the year from a PDF file name.

    Keyword arguments:
    text -- the text to extract the year from

    Returns: the year
    """
    year_pattern = re.compile(r"\d{4}", re.IGNORECASE)
    year = re.search(year_pattern, text)

    if year is not None:
        return int(year.group(0))
    else:
        raise Exception(f"Could not extract year from {text}")


def _is_header(row: list) -> bool:
    """
    Determine if a row is a header row.

    Keyword arguments:
    row -- the row to check

    Returns: True if the row is a header row, False otherwise
    """
    return row[0].strip().lower() == "company name"


def _is_clean_header(row: list) -> bool:
    """
    Return true for a header with a clean column name.

    Keyword arguments:
    row -- the rows to check

    Returns: true if the row is a clean header
    """
    return _is_header(row) and "Employees Affected" in row


def _extract_note(chars: list) -> str:
    """
    Extract a note from a PDF cell.

    Keyword arguments:
    chars -- the characters to extract the note from

    Returns: the note
    """
    text = pdfplumber.utils.extract_text(chars)

    # Split text into lines
    lines = text.split("\n")

    notes = []

    for line in lines:
        note_pattern = r"((UPDATE.*|WARN RESCINDED))+"
        note = re.search(note_pattern, line, re.IGNORECASE)
        if note:
            notes.append(_clean_text(note[0]))

    return " ".join(notes)


def _extract_company_name(chars: list) -> str:
    """
    Extract the company name from a PDF cell.

    Keyword arguments:
    chars -- the characters to extract the company name from

    Returns: the company name
    """
    text = pdfplumber.utils.extract_text(chars)

    # Split text into lines
    lines = text.split("\n")

    # We're assuming first line is always part of a company name
    company_name = _clean_text(lines[0])

    # Try to extract bold text in the cell
    bold_text = _clean_text(_extract_bold_text(chars))
    remaining_bold_text = bold_text.replace(company_name, "").strip()

    # Loop through all but first and last lines
    for index, line in enumerate(lines[1:-1]):
        line_text = _clean_text(line)

        # If line is bolded or doesn't match a pattern
        # it's probably part of the company name
        if remaining_bold_text.startswith(line_text) or (
            bold_text == "" and index < 1 and not _is_location(line_text)
        ):
            company_name += f" {line_text}"

            remaining_bold_text = remaining_bold_text.replace(line_text, "").strip()
        # The first time we hit a line that doesn't match our expectations,
        # we assume we've reached the end of the company name
        else:
            break

    return company_name


def _is_location(text: str) -> bool:
    """
    Determine if text is likely to be a location.

    Keyword arguments:
    text -- the text to check

    Returns: True if the text is likely to be a location, False otherwise
    """
    location_pattern = r"(^\d+|Highway|Hwy|Offshore|Statewide)"
    return re.match(location_pattern, text, re.IGNORECASE) is not None


def _extract_bold_text(chars: list) -> str:
    """
    Extract the bold text from a PDF cell.

    Keyword arguments:
    chars -- the list of characters in the cell

    Returns: the bold text
    """
    bold_chars = [char["text"] for char in chars if "Bold" in char["fontname"]]
    bold_text = "".join(bold_chars)
    return bold_text


def _clean_text(text: str) -> str:
    """
    Clean up text from a PDF cell.

    Keyword arguments:
    text -- the text to clean

    Returns: the cleaned text
    """
    # Replace None with an empty string
    if text is None:
        return ""

    # Standardize whitespace
    clean_text = re.sub(r"\s+", " ", text).strip()

    return clean_text


if __name__ == "__main__":
    scrape()



import csv
import logging
import os
import typing
from os.path import expanduser, join
from pathlib import Path

from .utils import get_url

logger = logging.getLogger(__name__)


class Cache:
    """Basic interface to save files to and fetch from cache.

    By default this will be: ~/.warn-scraper/cache

    With this directory, you can use partial paths to save or fetch
    file contents. State-specific files should generally be stored in a
    folder using the state's two-letter postal code.

    Example:
        Saving HTML to the cache::

            html = '<html><h1>Blob of HTML</h1></hmtl>'
            cache = Cache()
            cache.write('fl/2021_page_1.html', html)

        Retrieving pages from the cache::

            cache.files('fl')

    Args:
        path (str): Full path to cache directory. Defaults to WARN_ETL_DIR
            or, if env var not specified, $HOME/.warn-scraper/cache
    """

    def __init__(self, path=None):
        """Initialize a new instance."""
        self.root_dir = self._path_from_env or self._path_default
        self.path = path or str(Path(self.root_dir, "cache"))

    def exists(self, name):
        """Test whether the provided file path exists."""
        return Path(self.path, name).exists()

    def read(self, name):
        """Read text file from cache.

        Args:
            name (str): Partial name, relative to cache dir (eg. 'fl/2021_page_1.html')

        Returns:
            File content as string or error if file doesn't
        """
        path = Path(self.path, name)
        logger.debug(f"Reading from cache {path}")
        with open(path, newline="") as infile:
            return infile.read()

    def read_csv(self, name):
        """Read csv file from cache.

        Args:
            name (str): Partial name, relative to cache dir (eg. 'fl/2021_page_1.html')

        Returns:
            list of rows
        """
        path = Path(self.path, name)
        logger.debug(f"Reading CSV from cache {path}")
        with open(path) as fh:
            return list(csv.reader(fh))

    def download(
        self, name: str, url: str, encoding: typing.Optional[str] = None, **kwargs
    ) -> Path:
        """
        Download the provided URL and save it in the cache.

        Args:
            name (str): The path where the file will be saved. Can be a simple string like "ia/data.xlsx"
            url (str): The URL to download
            encoding (str): The encoding of the response. Optional.
            **kwargs: Additional arguments to pass to requests.get()

        Returns: The Path where the file was saved
        """
        # Request the URL
        logger.debug(f"Requesting {url}")
        with get_url(url, stream=True, **kwargs) as r:
            # If there's no encoding, set it
            if encoding:
                r.encoding = encoding
            elif r.encoding is None:
                r.encoding = "utf-8"

            # Open the local Path
            out_path = Path(self.path, name)
            out_path.parent.mkdir(parents=True, exist_ok=True)
            logger.debug(f"Writing to {out_path}")

            # Write out the file in little chunks
            with open(out_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        # Return the path
        return out_path

    def write(self, name, content):
        """Save file contents to cache.

        Typically, this should be a state-specific directory
        inside the cache folder.

        For example: ::

            $HOME/.warn-scraper/cache

        Provide file contents and the partial name for (relative to cache directory)
        where file should written. The partial file path can include additional
        directories (e.g. 'fl/2021_page_1.html'), which will be created if they
        don't exist.

        Example: ::

            cache.write("fl/page.html", html)

        Args:
            name (str): Partial name, relative to cache dir, where content should be saved.
            content (str): Any string content to save to file.
        """
        out = Path(self.path, name)
        out.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(f"Writing to cache {out}")
        with open(out, "w", newline="") as fh:
            fh.write(content)
        return str(out)

    def files(self, subdir=".", glob_pattern="*"):
        """
        Retrieve all files and folders in a subdir relative to cache dir.

        Examples:
            Given a cache dir such as $HOME/.warn-scraper/cache,
            you can: ::

                # Get all files and dirs in cache dir
                Cache().files()

                # Get files in specific subdir
                Cache().files('fl/')

                # Get all files of a specific type in a subdir
                Cache().files(subdir='fl/', glob_pattern='*.html')

        Args:
            subdir (str): Subdir inside cache to glob
            glob_pattern (str): Glob pattern. Defaults to all files in specified subdir ('*')
        """
        _dir = Path(self.path).joinpath(subdir)
        return [str(p) for p in _dir.glob(glob_pattern)]

    @property
    def _path_from_env(self):
        """Get the path where files will be saved."""
        return os.environ.get("WARN_ETL_DIR")

    @property
    def _path_default(self):
        """Get the default filesystem location of the cache."""
        return join(expanduser("~"), ".warn-scraper")



import typing
from pathlib import Path

import xlrd
from openpyxl import load_workbook

from .. import utils
from ..cache import Cache

__authors__ = [
    "palewire",
]
__tags__ = [
    "excel",
]


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Kentucky.

    Arguments:
    output_dir -- the Path were the result will be saved

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Get the latest workbook
    cache = Cache(cache_dir)
    latest_url = "https://kcc.ky.gov/WARN%20notices/WARN%20NOTICES%202022/WARN%20Notice%20Report%2001262022.xls"
    latest_path = cache.download("ky/latest.xls", latest_url)

    # Open it up
    latest_workbook = xlrd.open_workbook(latest_path)

    # Loop through the worksheets
    dict_list = []
    for sheet in latest_workbook.sheet_names():
        # Get the data as a list of lists
        worksheet = latest_workbook.sheet_by_name(sheet)
        sheet_list = []
        num_cols = worksheet.ncols
        for row_idx in range(0, worksheet.nrows):
            row = []
            for col_idx in range(0, num_cols):
                cell_obj = worksheet.cell(row_idx, col_idx)
                cell_type = worksheet.cell_type(row_idx, col_idx)
                # Carve out hack for the 2017 tab, which has the jobs as a date instead of an int
                if sheet == "2017" and col_idx == 5:
                    value = cell_obj.value
                # For all the rest of the dates
                elif cell_type == xlrd.XL_CELL_DATE:
                    value = xlrd.xldate.xldate_as_datetime(
                        cell_obj.value, latest_workbook.datemode
                    )
                # and then the rest of it
                else:
                    value = cell_obj.value
                row.append(value)
            sheet_list.append(row)

        # Convert to list of dicts
        headers = sheet_list[0]
        for row in sheet_list[1:]:
            d = {}
            for x, value in enumerate(row):
                d[headers[x]] = value
            dict_list.append(d)

    # Same thing for the archived file
    archive_url = "https://kcc.ky.gov/WARN%20notices/2017%20WARN%20docs/Kentucky%20%20WARN%20Report%20-Tracking%20Form%20%281998-2016%29.xlsx"
    archive_path = cache.download("ky/archive.xlsx", archive_url)

    # Open it up
    archive_workbook = load_workbook(filename=archive_path)

    # Loop through the worksheets
    for sheet in archive_workbook.worksheets:
        # Get the data as a list of lists
        sheet_list = parse_xlsx(sheet)

        # Convert to list of dicts
        headers = sheet_list[0]
        for row in sheet_list[1:]:
            d = {}
            for i, value in enumerate(row):
                d[headers[i]] = value
            dict_list.append(d)

    # Write out the results
    data_path = data_dir / "ky.csv"
    all_headers = list({value for dic in dict_list for value in dic.keys()})
    utils.write_dict_rows_to_csv(
        data_path, all_headers, dict_list, extrasaction="ignore"
    )

    # Pass it out
    return data_path


def parse_xlsx(worksheet) -> typing.List[typing.List]:
    """Parse the Excel xlsx file at the provided path.

    Args:
        worksheet: An openpyxl worksheet ready to parse

    Returns: List of values ready to write.
    """
    # Convert the sheet to a list of lists
    row_list = []
    for r in worksheet.rows:
        # Parse cells
        cell_list = [cell.value for cell in r]

        # Skip empty rows
        try:
            # A list with only empty cells will throw an error
            next(c for c in cell_list if c)
        except StopIteration:
            continue

        # Add to the master list
        row_list.append(cell_list)

    # Pass it back
    return row_list


if __name__ == "__main__":
    scrape()



import logging
import re
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "ydoc5212"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from South Dakota.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Open the cache
    cache = Cache(cache_dir)

    # Get the HTML
    url = "https://dlr.sd.gov/workforce_services/businesses/warn_notices.aspx"
    r = utils.get_url(url)
    html = r.text
    cache.write("sd/source.html", html)

    # Scrape out the data
    row_list = _parse_table(html)

    # Write out
    data_path = data_dir / "sd.csv"
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the CSV
    return data_path


def _parse_table(html) -> list:
    # Parse table
    soup = BeautifulSoup(html, "html.parser")
    table_list = soup.find_all("table")

    # We expect the first table to be there with our data
    assert len(table_list) > 0
    table = table_list[0]

    # Parse the cells
    row_list = []
    for row in table.find_all("tr"):
        cell_list = []
        for cell in row.find_all(["th", "td"]):
            cell = re.sub(r"\s+", " ", cell.text.strip())
            cell_list.append(cell)
        if not cell_list:
            continue
        row_list.append(cell_list)

    # Return it
    return row_list


if __name__ == "__main__":
    scrape()



import re
from datetime import datetime
from pathlib import Path

import pdfplumber
from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["palewire"]
__tags__ = [
    "html",
    "pdf",
]


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from South Carolina.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Get URL
    r = utils.get_url(
        "https://scworks.org/employer/employer-programs/at-risk-of-closing/layoff-notification-reports",
        verify=False,
    )
    html = r.text

    # Save it to the cache
    cache = Cache(cache_dir)
    cache.write("sc/source.html", html)

    # Parse out the PDF links
    soup = BeautifulSoup(html, "html.parser")
    link_list = soup.find_all("a")
    pdf_list = [a for a in link_list if "pdf" in a["href"]]

    # Pattern to find and extract data cells
    naics_re = re.compile("^[0-9]{5,6}$")
    date_re = re.compile("^[0-9]{1,2}/[0-9]{1,2}[/]{1,2}[0-9]{2}")
    jobs_re = re.compile("^[0-9]{1,4}$")

    current_year = datetime.now().year
    output_rows = []
    for pdf in pdf_list:
        pdf_year = int(pdf.text[:4].strip())
        cache_key = f"sc/{pdf_year}.pdf"
        if cache.exists(cache_key) and pdf_year < (current_year - 1):
            pdf_path = cache.path / cache_key
        else:
            pdf_path = cache.download(
                cache_key, f"https://scworks.org/{pdf['href']}", verify=False
            )

        # Open the PDF
        with pdfplumber.open(pdf_path) as pdf:

            # Loop through the pages
            for page in pdf.pages:

                # Pull out the table
                row_list = page.extract_table()

                # Skip empty pages
                if not row_list:
                    continue

                # Skip skinny and empty rows
                real_rows = []
                for row in row_list:
                    values = [v for v in row if v]
                    if len(values) < 4:
                        continue
                    real_rows.append(row)

                # Loop through each row in the table
                for row in real_rows:

                    # Clean values
                    cell_list = [_clean_cell(c) for c in row if _clean_cell(c)]

                    # Pluck out the values based on our regex
                    d = {}
                    for cell in cell_list:
                        if naics_re.search(cell):
                            d["naics"] = cell
                        elif date_re.search(cell):
                            d["date"] = cell
                        elif jobs_re.search(cell):
                            d["jobs"] = int(cell)

                    # If there haven't been at least two matches, it must be junk
                    if len(d) < 2:
                        continue

                    # The first one should be the company
                    d["company"] = cell_list[0]

                    # The second one should be the location
                    d["location"] = cell_list[1]

                    # Keep what we got
                    output_rows.append(d)

    # Write out the data to a CSV
    data_path = data_dir / "sc.csv"
    headers = ["company", "location", "date", "jobs", "naics"]
    utils.write_dict_rows_to_csv(data_path, headers, output_rows, extrasaction="ignore")

    # Return the Path to the CSV
    return data_path


def _clean_cell(cell):
    """Clean the value in the provided cell."""
    if cell is None:
        return None
    return cell.strip().replace("\n", "")


if __name__ == "__main__":
    scrape()



from pathlib import Path
from unittest.mock import patch

import pytest

from warn.cache import Cache

from .conftest import file_contents


def test_default_cache_dir():
    """Override the output of the expanduser method."""
    to_patch = "warn.cache.expanduser"
    with patch(to_patch) as mock_func:
        mock_func.return_value = "/Users/you"
        cache = Cache()
        assert cache.path == "/Users/you/.warn-scraper/cache"


def test_custom_cache_path(tmpdir):
    """Test if the provided path matches the cache's path."""
    from warn.cache import Cache

    cache = Cache(tmpdir)
    assert tmpdir == cache.path


def test_write(tmpdir):
    """Test writing to the cache."""
    from warn.cache import Cache

    cache = Cache(tmpdir)
    content = "<h1>some content</h1>"
    file_path = "fl/2021_page_1.html"
    outfile = cache.write(file_path, content)
    scrape_dir = tmpdir.join("fl")
    files = [f.basename for f in scrape_dir.listdir()]
    assert "2021_page_1.html" in files
    actual_contents = file_contents(outfile)
    assert actual_contents == content


@pytest.mark.usefixtures("create_cache_dir", "copy_html_to_cache")
def test_files(cache_dir):
    """Test the file fetching function of the cache."""
    from warn.cache import Cache

    cache = Cache(path=cache_dir)
    # Base cache dir only contains fl/ dir
    assert cache.files()[0].endswith("/fl")
    # HTML files are stored in fl/ directory
    actual = [str(p) for p in Path(cache_dir, "fl").glob("*.html")]
    assert cache.files(subdir="fl/") == actual


@pytest.mark.usefixtures("create_cache_dir", "copy_html_to_cache")
def test_exists(cache_dir):
    """Test the exists method of the cache."""
    from warn.cache import Cache

    cache = Cache(path=cache_dir)
    assert cache.exists("fl/2021_page_1.html")


@pytest.mark.usefixtures("create_cache_dir", "copy_html_to_cache")
def test_read(cache_dir):
    """Test reading from the cache."""
    from warn.cache import Cache

    cache = Cache(path=cache_dir)
    html = cache.read("fl/2021_page_1.html").strip()
    assert html == "<html><h1>2021 page 1</h1></html>"



import logging
import re
from datetime import datetime
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "shallotly"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Missouri.

    NOTES for data cleaning:
    - 2019 and 2020 page has duplicate data
    - 2017 date format is different

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Set the cache
    cache = Cache(cache_dir)

    # Get the range of years we're after
    today = datetime.today()
    current_year = today.year
    year_range = list(range(2015, current_year + 1))
    year_range.reverse()

    # Download them all
    html_list = []
    for year in year_range:

        # Set the URL, with a hack for 2020 and 2022
        if year == 2020:
            url = "https://jobs.mo.gov/content/2020-missouri-warn-notices"
        elif year == 2022:
            url = "https://jobs.mo.gov/content/2022-warn-notices"
        else:
            url = f"https://jobs.mo.gov/warn{year}"

        # Read from cache if available and not this year or the year before
        cache_key = f"mo/{year}.html"
        if cache.exists(cache_key) and year < current_year - 1:
            html = cache.read(cache_key)
        else:
            # Otherwise, go request it
            r = utils.get_url(url)
            html = r.text
            # Save it to the cache
            cache.write(cache_key, html)

        # Add it to the list
        html_list.append(html)

    # Parse them all
    output_rows = []
    for i, html in enumerate(html_list):
        soup = BeautifulSoup(html, "html5lib")

        # Pull out the table
        table_list = soup.find_all("table")
        assert len(table_list) > 0
        table = table_list[0]

        # Get all rows
        row_list = table.find_all("tr")

        # If it's not the first page, slice off the header
        if i > 0:
            row_list = row_list[1:]

        # Loop through all the rows
        year_rows = []
        for row in row_list:
            # Get the cells
            cell_list = row.find_all(["td", "th"])

            # Clean them up
            cell_list = [_clean_text(c.text) for c in cell_list]

            if len(cell_list) < 9:  # to account for the extra column in 2021
                cell_list.insert(2, "")

            # Pass them out
            year_rows.append(cell_list)

        # pop "Total" row
        year_rows.pop(len(year_rows) - 1)

        # Add to master list
        output_rows.extend(year_rows)

    # Set the export path
    data_path = data_dir / "mo.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the file
    return data_path


def _clean_text(text):
    """Clean up the provided HTML snippet."""
    if text is None:
        return ""
    text = re.sub(r"\n", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text.strip()


if __name__ == "__main__":
    scrape()



import logging
from datetime import datetime
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Nebraska.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Open the cache
    cache = Cache(cache_dir)

    current_year = datetime.now().year
    year_range = range(2010, current_year + 1)

    # Scrape rows
    output_rows = []
    for year in year_range:
        # Get WARN page
        warn_url = (
            f"https://dol.nebraska.gov/LayoffServices/WARNReportData/?year={year}"
        )
        warn_key = f"ne/warn-{year}.html"

        # Read from cache if available and not this year or the year before
        if cache.exists(warn_key) and year < current_year - 1:
            warn_html = cache.read(warn_key)
        else:
            warn_r = utils.get_url(warn_url)
            warn_html = warn_r.text
            cache.write(warn_key, warn_html)

        # Parse the table
        warn_headers = ["Date", "Company", "Jobs Affected", "City", "Location"]
        warn_rows = _parse_table(warn_html, warn_headers)

        # Add it to the big list
        output_rows.extend(warn_rows)

        # Do the same for the layoffs page
        layoff_url = f"https://dol.nebraska.gov/LayoffServices/LayoffAndClosureReportData/?year={year}"
        layoff_key = f"ne/layoff-{year}.html"

        # Read from cache if available and not this year or the year before
        if cache.exists(layoff_key) and year < current_year - 1:
            layoff_html = cache.read(layoff_key)
        else:
            page = utils.get_url(layoff_url)
            layoff_html = page.text
            cache.write(layoff_key, layoff_html)

        # Parse the table
        layoff_headers = [
            "Date",
            "Company",
            "Type",
            "Jobs Affected",
            "City",
            "Location",
        ]
        layoff_rows = _parse_table(layoff_html, layoff_headers)

        # Add it to the big list
        output_rows.extend(layoff_rows)

    # Write out the results
    data_path = data_dir / "ne.csv"
    utils.write_dict_rows_to_csv(
        data_path, layoff_headers, output_rows, extrasaction="ignore"
    )

    # Return the path to the CSV
    return data_path


def _parse_table(html, headers) -> list:
    # Parse table
    soup = BeautifulSoup(html, "html.parser")
    table_list = soup.find_all("table")

    # We expect the first table to be there with our data
    assert len(table_list) > 0
    table = table_list[0]

    # Parse the cells
    row_list = []
    for row in table.find_all("tr"):
        cell_list = row.find_all("td")
        if not cell_list:
            continue
        cell_dict = {headers[i]: c.text.strip() for i, c in enumerate(cell_list)}
        row_list.append(cell_dict)

    # Return it
    return row_list


if __name__ == "__main__":
    scrape()



import re
from pathlib import Path

import pytest

from warn.platforms import JobCenterSite


@pytest.fixture
def ks_site(cache_dir):
    """Create a fixture from the Kansas site."""
    url = "https://www.kansasworks.com/search/warn_lookups"
    return JobCenterSite("KS", url, cache_dir)


@pytest.fixture
def ok_site(cache_dir):
    """Create a fixture from the Oklahoma site."""
    url = "https://okjobmatch.com/search/warn_lookups"
    return JobCenterSite("OK", url, cache_dir)


@pytest.mark.vcr()
def test_scrape_integration(ok_site):
    """Test a scraper."""
    # should scrape by start and end dates
    results_pages, data = ok_site.scrape(
        start_date="2021-01-01",
        end_date="2021-02-03",
        detail_pages=True,
    )
    assert list(results_pages.keys()) == [1]
    assert isinstance(results_pages[1], str)
    assert len(data) == 2
    first = data[0]
    # Basics from search result pages
    assert first["employer"] == "Exterran Energy Solutions"
    assert first["city"] == "Broken Arrow"
    assert first["zip"] == "74014"
    assert first["lwib_area"] == "12 - TulsaRegion"
    assert first["notice_date"] == "Jan 13, 2021"
    assert first["warn_type"] == "WARN"
    assert first["detail_page_url"] == "https://okjobmatch.com/search/warn_lookups/780"
    # Additional data from record detail page
    details = first["detail"]
    assert details["record_number"] == "780"
    assert details["company_name"] == "Exterran Energy Solutions"
    assert details["address"] == "20602 E. 81st Street\nBroken Arrow, Oklahoma 74014"
    assert details["notice_date"] == "Jan 13, 2021"
    assert details["number_of_employees_affected"] == "0"
    assert "html" in details.keys()


@pytest.mark.vcr()
def test_no_results(ks_site):
    """Test an instance where there are no results."""
    # The dates below should span two pages (just barely).
    # Skip detail pages to minimize fixture size.
    results_pages, data = ks_site.scrape(
        start_date="1997-01-01",
        end_date="1997-12-31",
        detail_pages=False,
    )
    assert data == []
    assert results_pages == {}


@pytest.mark.vcr()
def test_missing_detail_page_values(ks_site):
    """Test a case where detail pages are missing one or more field values."""
    results_pages, data = ks_site.scrape(
        start_date="2020-07-31",
        end_date="2020-07-31",
        detail_pages=True,
    )
    detail = data[0]["detail"]
    assert data[0]["employer"] == "Spirit AeroSystems, Inc."
    assert detail["company_name"] == "Spirit AeroSystems, Inc."
    # This record has a malformed date
    assert detail["notice_date"] == "Jul 31, 2020"
    assert detail["number_of_employees_affected"] == "1100"
    # Record is missing the address field
    assert detail["address"] == ""


@pytest.mark.vcr()
def test_paged_results(ok_site):
    """Test a case with paginated results."""
    # The dates below should span two pages (just barely).
    # Skip detail pages to minimize fixture size.
    results_pages, data = ok_site.scrape(
        start_date="2020-01-01",
        end_date="2020-04-30",
        detail_pages=False,
    )
    assert len(data) == 42
    first = data[0]
    last = data[-1]
    assert first["employer"] == "Vallourec Star, LP"
    assert last["employer"] == "Uncle Julio's"


@pytest.mark.vcr()
def test_cached_search_results(tmp_path):
    """Test a case when the cache should be used for search results."""
    url = "https://okjobmatch.com/search/warn_lookups"
    cache_dir = str(tmp_path.joinpath("ok"))
    site = JobCenterSite("OK", url, cache_dir=cache_dir)
    # The dates below should span two pages (just barely).
    # Skip detail pages to minimize fixture size.
    results_pages, data = site.scrape(
        start_date="2020-01-01",
        end_date="2020-04-15",
        detail_pages=False,
    )
    expected = ["2020-01-01_2020-04-15_page1.html", "2020-01-01_2020-04-15_page2.html"]
    files = [f.name for f in Path(cache_dir).glob("**/*.html")]
    assert len(files) == 2
    assert sorted(files) == sorted(expected)


@pytest.mark.vcr()
def test_cached_detail_pages(tmp_path):
    """Test a case when the cache should be used for detail pages."""
    url = "https://okjobmatch.com/search/warn_lookups"
    cache_dir = str(tmp_path.joinpath("ok"))
    site = JobCenterSite("OK", url, cache_dir=cache_dir)
    # The dates below should span two pages (just barely).
    # Skip detail pages to minimize fixture size.
    results_pages, data = site.scrape(
        start_date="2021-01-01",
        end_date="2021-02-03",
        detail_pages=True,
    )
    assert len(data) == 2
    files = [f.name for f in Path(cache_dir).glob("**/*.html")]
    assert len(files) == 3
    # record files returned by glob resemble 1234.html (no records prefix)
    record_files = [f for f in files if re.match(r"\d{1,4}.html", f)]
    assert len(record_files) == 2
    assert Path(cache_dir, "records").exists()
    assert Path(cache_dir, "search_results").exists()



import logging
import os
import re
from datetime import datetime
from pathlib import Path
from typing import Optional

import pdfplumber
from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["chriszs"]
__tags__ = ["pdf"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from New Mexico.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Fire up the cache
    cache = Cache(cache_dir)

    # Download the root page
    base_url = "https://www.dws.state.nm.us/"
    url = f"{base_url}Rapid-Response"
    page = utils.get_url(url)
    html = page.text

    # Save it to the cache
    state_code = "nm"
    cache_key = f"{state_code}/Rapid-Response.html"
    cache.write(cache_key, html)

    # Parse out the PDF links
    document = BeautifulSoup(html, "html.parser")
    links = document.find_all("a")
    pdf_urls = [
        f"{base_url}{link['href']}"
        for link in links
        if "WARN" in link.get("href", "") and link.get("href", "").endswith(".pdf")
    ]

    output_rows = []

    for pdf_index, pdf_url in enumerate(pdf_urls):
        file_name = os.path.basename(pdf_url)
        cache_key = f"{state_code}/{file_name}"
        year = _extract_year(file_name)
        current_year = datetime.now().year
        if cache.exists(cache_key) and year is not None and year < current_year - 1:
            pdf_path = Path(cache_dir, cache_key)
        else:
            pdf_path = cache.download(cache_key, pdf_url)

        with pdfplumber.open(pdf_path) as pdf:
            for page_index, page in enumerate(pdf.pages):
                rows = page.extract_table()

                # Loop through the rows
                for row_index, row in enumerate(rows):
                    # Skip headers on all but first page of first PDF
                    if pdf_index > 0 and row_index == 0:
                        logger.debug(
                            f"Skipping header row on PDF {pdf_index+1} page {page_index+1}"
                        )
                        continue

                    # Extract data
                    output_row = [_clean_text(cell) for cell in row]

                    # Write row
                    if any([cell != "" for cell in output_row]):
                        output_rows.append(output_row)

    # Write out to CSV
    data_path = data_dir / f"{state_code}.csv"
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path
    return data_path


def _clean_text(text: str) -> str:
    """
    Clean up text from a PDF cell.

    Keyword arguments:
    text -- the text to clean

    Returns: the cleaned text
    """
    # Replace None with an empty string
    if text is None:
        return ""

    # Standardize whitespace
    return re.sub(r"\s+", " ", text)


def _extract_year(text: str) -> Optional[int]:
    """
    Extract the year from a string.

    Keyword arguments:
    text -- the string to extract the year from

    Returns: the year
    """
    match = re.search(r"\d{4}", text)

    if match is not None:
        return int(match.group(0))

    return None


if __name__ == "__main__":
    scrape()



import logging
from pathlib import Path

import click

from . import Runner, utils


@click.command()
@click.argument("scrapers", nargs=-1)
@click.option(
    "--data-dir",
    default=utils.WARN_DATA_DIR,
    type=click.Path(),
    help="The Path were the results will be saved",
)
@click.option(
    "--cache-dir",
    default=utils.WARN_CACHE_DIR,
    type=click.Path(),
    help="The Path where results can be cached",
)
@click.option(
    "--delete/--no-delete",
    default=False,
    help="Delete generated files from the cache",
)
@click.option(
    "--log-level",
    "-l",
    default="INFO",
    type=click.Choice(
        ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"), case_sensitive=False
    ),
    help="Set the logging level",
)
def main(
    scrapers: list,
    data_dir: Path,
    cache_dir: Path,
    delete: bool,
    log_level: str,
):
    """
    Command-line interface for downloading WARN Act notices.

    SCRAPERS -- a list of one or more postal codes to scrape. Pass `all` to scrape all supported states and territories.
    """
    # Set higher log-level on third-party libs that use DEBUG logging,
    # In order to limit debug logging to our library
    logging.getLogger("urllib3").setLevel(logging.ERROR)
    logging.getLogger("pdfminer").setLevel(logging.WARNING)

    # Local logging config
    logging.basicConfig(level=log_level, format="%(asctime)s - %(name)s - %(message)s")
    logger = logging.getLogger(__name__)

    # Runner config
    data_dir = Path(data_dir)
    cache_dir = Path(cache_dir)
    runner = Runner(data_dir, cache_dir)

    # Delete files, if asked
    if delete:
        logger.info("Deleting files generated from previous scraper run.")
        runner.delete()

    # If the user has asked for all states, give it to 'em
    if "all" in scrapers:
        scrapers = utils.get_all_scrapers()

    # Loop through the states
    for scrape in scrapers:
        # Try running the scraper
        runner.scrape(scrape)


if __name__ == "__main__":
    main()



from pathlib import Path

from warn.platforms.job_center.utils import scrape_state

from .. import utils

__authors__ = ["zstumgoren", "Dilcia19"]
__tags__ = [
    "jobcenter",
]


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
    use_cache: bool = True,
) -> Path:
    """
    Scrape data from Arizona.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)
    use_cache -- a Boolean indicating whether the cache should be used (default True)

    Returns: the Path where the file is written
    """
    output_csv = data_dir / "az.csv"
    search_url = "https://www.azjobconnection.gov/search/warn_lookups"

    # Date chosen based on manual research
    stop_year = 2010

    # Use cache for years before current and prior year
    scrape_state(
        "AZ", search_url, output_csv, stop_year, cache_dir, use_cache=use_cache
    )

    return output_csv


if __name__ == "__main__":
    scrape()



import logging
from pathlib import Path

from .. import utils
from ..cache import Cache

__authors__ = ["chriszs"]
__tags__ = ["html", "excel"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Illinois.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    cache = Cache(cache_dir)

    state_code = "il"

    # Get the file
    url = "https://apps.illinoisworknet.com/iebs/api/public/export?search=&layoffTypes=&trade=0&dateReportedStart=Invalid%20Date&dateReportedEnd=Invalid%20Date&statuses=4&reasons=&eventCauses=&naicsCodes=1&naicIndustries=&naics=&unionsInvolved=0&geolocation=1&cities=&counties=&lwias=&includeAdditionalLwias=false&edrs=&lat=0&lng=0&distance=.5&memberType=1&users=&accessList=&bookmarked=false"
    file_path = cache.download(f"{state_code}/export.xlsx", url)

    # Parse it
    row_list = utils.parse_excel(file_path)

    # Write out the results
    data_path = data_dir / f"{state_code}.csv"
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the CSV
    return data_path


if __name__ == "__main__":
    scrape()



import logging
import re
import shutil
from pathlib import Path

import pdfplumber
import requests
from bs4 import BeautifulSoup
from openpyxl import load_workbook

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "ydoc5212"]
__tags__ = ["html", "pdf", "excel"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from California.

    Compiles a single CSV for CA using historical PDFs and an Excel file for the current fiscal year.

    Only regenerates the CSV if a PDF or the Excel file have changed.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    output_csv = data_dir / "ca.csv"
    # Set up cache dir for state
    cache_state = Path(cache_dir, "ca")
    cache_state.mkdir(parents=True, exist_ok=True)
    # Initially write to a temp file in cache_dir before
    # over-writing prior output_csv, so we can use append
    # mode while avoiding data corruption if script errors out
    temp_csv = cache_state / "ca_temp.csv"
    # Create Cache instance for downstream operations
    cache = Cache(cache_dir)
    # Update pdfs and Excel files
    files_have_changed = _update_files(cache)
    output_headers = [
        "notice_date",
        "effective_date",
        "received_date",
        "company",
        "city",
        "num_employees",
        "layoff_or_closure",
        "county",
        "address",
        "source_file",
    ]
    if files_have_changed:
        logger.info("One or more source data files have changed")
        logger.info("Extracting Excel data for current fiscal year")
        wb_path = cache.files(subdir="ca", glob_pattern="*.xlsx")[0]
        excel_data = _extract_excel_data(wb_path)
        # Write mode when processing Excel
        utils.write_dict_rows_to_csv(
            temp_csv, output_headers, excel_data, mode="w", extrasaction="ignore"
        )
        logger.info("Extracting PDF data for prior years")
        for pdf in cache.files(subdir="ca", glob_pattern="*.pdf"):
            logger.info(f"Extracting data from {pdf}")
            data = _extract_pdf_data(pdf)
            # Append mode when processing PDFs
            utils.write_dict_rows_to_csv(
                temp_csv, output_headers, data, mode="a", extrasaction="ignore"
            )
        # If all went well, copy temp csv over pre-existing output csv
        utils.create_directory(output_csv, is_file=True)
        shutil.copy2(temp_csv, output_csv)
    return output_csv


def _update_files(cache):
    files_have_changed = False
    # Create lookup of pre-existing PDF files and their size
    files = {}
    for local_file in cache.files(subdir="ca/"):
        fname = local_file.split("/")[-1]
        extension = fname.split(".")[-1]
        if extension in ["pdf", "xlsx"]:
            files[fname] = Path(local_file).stat().st_size
    # Download file if it has changed or not present.
    links = _get_file_links()
    for link in links:
        file_name = link["url"].split("/")[-1]
        download_status = False
        # If file doesn't exist, update download status
        try:
            local_size = files[file_name]
        except KeyError:
            download_status = True
            local_size = None
        # If size doesn't match, update download status
        if local_size != link["size"]:
            download_status = True
        if download_status is True:
            files_have_changed = True
            cache.download(f"ca/{file_name}", link["url"])
    # Delete local files whose names don't match
    # data files on remote site, in order to guard against
    # duplicates if the source agency renames files
    for obsolete_file in _obsolete_local_files(files, links):
        files_have_changed = True
        logger.info(
            f"Deleting local file no longer present on source site: {obsolete_file}"
        )
        Path(cache.path, f"ca/{obsolete_file}").unlink()
    return files_have_changed


def _get_file_links():
    """Get links to historical PDFs and the Excel file."""
    logger.info("Getting metadata for data files")
    base_url = "https://edd.ca.gov/Jobs_and_Training"
    home_page = f"{base_url}/Layoff_Services_WARN.htm"
    html = utils.get_url(home_page).text
    soup = BeautifulSoup(html, "html.parser")
    links = []
    for link in soup.find_all("a"):
        relative_file_url = link.attrs.get("href", "").strip()
        if _is_warn_report_link(relative_file_url):
            file_url = f"{base_url}/{relative_file_url}"
            meta = _get_file_metadata(file_url)
            links.append(meta)
    return links


def _is_warn_report_link(url):
    return True if re.search(r"warn[-_]?report", url, re.I) else False


def _get_file_metadata(url):
    return {"url": url, "size": int(requests.head(url).headers["Content-Length"])}


def _extract_excel_data(wb_path):
    wb = load_workbook(filename=wb_path)
    # Get the only worksheet
    ws = wb.worksheets[0]
    rows = [row for row in ws.rows]
    # Throw away initial rows until we reach first data row
    while True:
        row = rows.pop(0)
        first_cell = row[0].value.strip().lower()
        if first_cell.startswith("county"):
            break
    payload = []
    for row in rows:
        first_cell = row[0].value.strip().lower()
        # Exit if we've reached summary row at bottom
        if first_cell == "report summary":
            break
        # Spreadsheet contains merged cells so index
        # positions below are not sequential
        data = {
            "county": row[0].value.strip(),
            "notice_date": _convert_date(row[1].value),
            "received_date": _convert_date(row[2].value),
            "effective_date": _convert_date(row[4].value),
            "company": row[5].value.strip(),
            "layoff_or_closure": row[8].value.strip(),
            "num_employees": row[10].value,
            "address": row[12].value.strip(),
            "source_file": wb_path.split("/")[-1],
        }
        payload.append(data)
    return payload


def _convert_date(dt):
    return dt.strftime("%m/%d/%Y")


def _extract_pdf_data(pdf_path):
    headers = [
        "notice_date",
        "effective_date",
        "received_date",
        "company",
        "city",
        "county",
        "num_employees",
        "layoff_or_closure",
        "source_file",
    ]
    data = []
    with pdfplumber.open(pdf_path) as pdf:
        for idx, page in enumerate(pdf.pages):
            # All pages pages except last should have a single table
            # Last page has an extra summary table, but indexing
            # for the first should avoid grabbing the summary data
            rows = page.extract_tables()[0]
            # Remove header row on first page
            # and update the standardized "headers" var if the source
            # data has no county field, as in the case of
            # files covering 07/2016-to-06/2017 fiscal year and earlier
            if idx == 0:
                raw_header = rows.pop(0)
                raw_header_str = "-".join([col.strip().lower() for col in raw_header])
                if "county" not in raw_header_str:
                    headers.remove("county")
            # Skip if it's a summary table (this happens
            # when summary is only table on page, as in 7/2019-6/2020)
            first_cell = rows[0][0].strip().lower()
            if "summary" in first_cell:
                continue
            for row in rows:
                data_row = dict(zip(headers, row))
                # Data clean-ups
                data_row.update(
                    {
                        "effective_date": data_row["effective_date"].replace(" ", ""),
                        "received_date": data_row["received_date"].replace(" ", ""),
                        "source_file": pdf_path.split("/")[-1],
                    }
                )
                data.append(data_row)
    return data


def _obsolete_local_files(pdfs, links):
    pdfs_uniq = set(pdfs.keys())
    remote_files = {link["url"].split("/")[-1] for link in links}
    return pdfs_uniq - remote_files


if __name__ == "__main__":
    scrape()



import logging
import re
from pathlib import Path

from bs4 import BeautifulSoup
from openpyxl import load_workbook

from .. import utils
from ..cache import Cache

__authors__ = ["Dilcia19", "ydoc5212"]
__tags__ = ["html", "excel", "historical"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Texas.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Set up the cache
    cache = Cache(cache_dir)

    # Get the root URL
    url = "https://www.twc.texas.gov/businesses/worker-adjustment-and-retraining-notification-warn-notices#warnNotices"
    page = utils.get_url(url)
    html = page.text

    # Cache it
    cache.write("tx/source.html", html)

    # Get all the Excel links
    soup = BeautifulSoup(page.text, "html5lib")
    link_list = soup.find_all("a", href=re.compile("^/files/news/warn-act-listings-"))

    # Clean up the links and filter 'em down
    href_list = []
    for link in link_list:
        # Extract year
        href = link.get("href")
        # Only keep links after 2019 since our historical document covers 2018 and before
        year = _get_year(href)
        if year >= 2019:
            href_list.append(href)

    # Loop through the links we want to download
    row_list = []
    for ihref, href in enumerate(href_list):

        # get each url from the HTML links we found
        data_url = f"https://www.twc.texas.gov{href}"

        # download the excel file
        year = _get_year(href)
        ext = _get_ext(href)
        excel_path = cache.download(f"tx/{year}{ext}", data_url)

        # Open it up
        workbook = load_workbook(filename=excel_path)

        # Get the first sheet
        worksheet = workbook.worksheets[0]

        # Convert the sheet to a list of lists
        for irow, row in enumerate(worksheet.rows):

            # Skip headers after the first workbook
            if ihref > 0 and irow == 0:
                continue
            cell_list = [cell.value for cell in row]

            # Skip empty rows
            if cell_list[0] is None:
                continue

            # Add what's left to the pile
            row_list.append(cell_list)

    # Get historical URL
    historical_url = (
        "https://storage.googleapis.com/bln-data-public/warn-layoffs/tx_historical.xlsx"
    )
    excel_path = cache.download("tx/historical.xlsx", historical_url)

    # Open it up
    workbook = load_workbook(filename=excel_path)

    # Get the first sheet
    worksheet = workbook.worksheets[0]

    # Convert the sheet to a list of lists
    for i, row in enumerate(worksheet.rows):
        # Skip header
        if i == 0:
            continue

        # Trim down to only the columns in the scrape, so they match
        select_columns = [
            row[8],  # NOTICE_DATE
            row[0],  # JOB_SITE_NAME
            row[2],  # COUNTY_NAME
            row[5],  # WDA_NAME
            row[6],  # TOTAL_LAYOFF_NUMBER
            row[7],  # LayOff_Date
            row[11],  # WFDD_RECEIVED_DATE
            row[1],  # CITY_NAME
        ]

        # Tack 'em on
        cell_list = [c.value for c in select_columns]
        row_list.append(cell_list)

    # Set the export path
    data_path = data_dir / "tx.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the file
    return data_path


def _get_year(url: str) -> int:
    """Plucks the year from the provided URL."""
    filename_regex = re.match(r".*-(.{4})(\..*)$", url, re.I)
    assert filename_regex is not None
    year_str = filename_regex.group(1)[-4:]
    return int(year_str)


def _get_ext(url: str) -> str:
    """Plucks the file extension from the provided URL."""
    filename_regex = re.match(r".*-(.{4})(\..*)$", url, re.I)
    assert filename_regex is not None
    return filename_regex.group(2)


if __name__ == "__main__":
    scrape()



import logging
import re
import uuid
from datetime import datetime
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "shallotly"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Washington D.C.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Set the cache
    cache = Cache(cache_dir)

    # Get the root page
    today = datetime.today()
    current_year = today.year
    url = f"https://does.dc.gov/page/industry-closings-and-layoffs-warn-notifications-{current_year}"
    r = utils.get_url(url)
    r.encoding = "utf-8"
    root_html = r.text

    # Save it to the cache
    cache.write(f"dc/{current_year}.html", root_html)

    # Parse the list of links
    soup = BeautifulSoup(root_html, "html5lib")
    table_list = soup.find_all("div", {"class": "field-items"})
    assert len(table_list) > 0

    # Grab all the links
    link_table = table_list[0]
    link_lookup = {_extract_year(a.string): a["href"] for a in link_table.find_all("a")}

    # As documented in #238, the page for 2014 is missing. We're
    # testing whether the URL for the 2019 page is the same as the
    # the 2014 link currently points to and scraping an archived copy
    # from 2017 instead.
    if link_lookup.get("2014") == link_lookup.get("2018"):
        logger.warning("2014 link is the same as 2018 link, using archived 2014")
        link_lookup[
            "2014"
        ] = "https://web.archive.org/web/20170210010137/http://does.dc.gov/page/industry-closings-and-layoffs-warn-notifications-closure%202014"

    # Download them all
    html_list = [
        root_html,
    ]
    for href in link_lookup.values():

        # Request the HTML
        r = utils.get_url(href)
        r.encoding = "utf-8"
        html = r.text

        # Save it to the cache
        cache_key = uuid.uuid5(uuid.NAMESPACE_URL, href)
        cache.write(f"dc/{cache_key}.html", html)

        # Add it to the list
        html_list.append(html)

    # Parse them all
    output_rows = []
    for i, html in enumerate(html_list):
        soup = BeautifulSoup(html, "html5lib")

        # Pull out the table
        table_list = soup.find_all("table")
        assert len(table_list) > 0
        table = table_list[0]

        # Get all rows
        row_list = table.find_all("tr")

        # If it's not the first page, slice off the header
        if i > 0:
            row_list = row_list[1:]

        # Loop through all the rows
        for row in row_list:
            # Get the cells
            cell_list = row.find_all(["td", "th"])

            # Clean them up
            cell_list = [_clean_text(c.text) for c in cell_list]

            # Add to the list if any cell in the row has data
            # (filters out empty rows)
            if any(cell_list):
                output_rows.append(cell_list)

    # Set the export path
    data_path = data_dir / "dc.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the file
    return data_path


def _clean_text(text):
    """Clean up the provided HTML snippet."""
    if text is None:
        return ""
    text = re.sub(r"\n", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text.strip()


def _extract_year(text):
    """Extract the year from the string."""
    if text is None:
        return None
    return re.sub(r"\D", "", text)


if __name__ == "__main__":
    scrape()



import logging
import random
import re
from pathlib import Path

import pdfplumber

from .. import utils
from ..cache import Cache

__authors__ = ["chriszs"]
__tags__ = ["pdf"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Idaho.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Create the URL of the source PDF
    base_url = "https://www.labor.idaho.gov/dnn/Portals/0/Publications/"
    file_name = "WARNNotice.pdf"
    # There's a numeric parameter called v on this PDF URL that updates
    # from time to time. Suspect this is a cache-buster. We're using a
    # random number instead.
    min_cache_buster = 0
    max_cache_buster = 10000000000
    cache_buster = random.randrange(min_cache_buster, max_cache_buster)
    url = f"{base_url}{file_name}?v={cache_buster}"

    # Download the PDF with verify=False because
    # there's a persistent cert error we're working around.
    cache = Cache(cache_dir)
    state_code = "id"
    cache_key = f"{state_code}/{file_name}"
    pdf_file = cache.download(cache_key, url, verify=False)

    # Loop through the PDF pages and scrape out the data
    output_rows: list = []
    with pdfplumber.open(pdf_file) as pdf:
        for index, page in enumerate(pdf.pages):
            rows = page.extract_table()
            output_rows += _clean_table(rows, index)

    # Write out the data to a CSV
    data_path = data_dir / f"{state_code}.csv"
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the CSV
    return data_path


def _clean_table(rows: list, page_index: int) -> list:
    """
    Clean up a table from a PDF.

    Keyword arguments:
    rows -- the rows of the table
    page_index -- the index of the page

    Returns: a list of lists, where each inner list is a row in the table
    """
    output_rows: list = []

    for row_index, row in enumerate(rows):
        output_row = []
        for col_index, column in enumerate(row):
            clean_text = _clean_text(column)

            # If cell is empty, copy from the cell above it
            # to deal with merged cells. Except for number of employees,
            # which is effectively a total for all locations in the merged cell
            # and which we don't want a data user to double count.
            if (
                _is_empty(clean_text)
                and _column_exists_in_prior_row(output_rows, row_index, col_index)
                and "No. of Employees"
                not in _column_name_from_index(output_rows, col_index)
            ):
                clean_text = output_rows[row_index - 1][col_index]

            output_row.append(clean_text)

        output_rows.append(output_row)

    # Only include the header on the first page
    if page_index != 0:
        return output_rows[1:]

    return output_rows


def _is_empty(text: str) -> bool:
    """
    Determine if a cell is empty.

    Keyword arguments:
    text -- the text to check

    Returns: True if the cell is empty, False otherwise
    """
    return text == ""


def _column_exists_in_prior_row(
    output_rows: list, row_index: int, col_index: int
) -> bool:
    """
    Determine if a column exists in the prior row.

    Keyword arguments:
    row_index -- the index of the row
    col_index -- the index of the column
    output_rows -- the output rows

    Returns: True if the column exists, False otherwise
    """
    return row_index > 0 and col_index < len(output_rows[row_index - 1])


def _column_name_from_index(output_rows: list, col_index: int) -> str:
    """
    Determine the column name from the column index.

    Keyword arguments:
    col_index -- the index of the column
    output_rows -- the output rows

    Returns: the column name
    """
    return output_rows[0][col_index]


def _clean_text(text: str) -> str:
    """
    Clean up text from a PDF cell.

    Keyword arguments:
    text -- the text to clean

    Returns: the cleaned text
    """
    if text is None:
        return ""
    # Collapse newlines
    partial = re.sub(r"\n", " ", text)
    # Standardize whitespace
    return re.sub(r"\s+", " ", partial)


if __name__ == "__main__":
    scrape()



import logging
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "shallotly", "palewire"]
__tags__ = ["html", "excel"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Iowa.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Go to the page
    url = "https://www.iowaworkforcedevelopment.gov/worker-adjustment-and-retraining-notification-act"
    r = utils.get_url(url)
    html = r.text

    # Save it to the cache
    cache = Cache(cache_dir)
    cache.write("ia/source.html", html)

    # Parse out the Excel link
    soup = BeautifulSoup(html, "html.parser")
    excel_url = soup.find("a", {"title": "WARN Log Excel File"})["href"]

    # Download the Excel file
    excel_path = cache.download("ia/source.xlsx", excel_url)

    # Parse it
    row_list = utils.parse_excel(excel_path)

    # Get historic file
    historic_url = "https://www.iowaworkforcedevelopment.gov/sites/search.iowaworkforcedevelopment.gov/files/documents/2018/WARN_20180503.xlsx"
    historic_excel_path = cache.download("ia/historic.xlsx", historic_url)

    # Parse it, minus the header
    row_list += utils.parse_excel(historic_excel_path, keep_header=False)

    # Set the export path
    data_path = data_dir / "ia.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the file
    return data_path


if __name__ == "__main__":
    scrape()



import logging
from pathlib import Path

from openpyxl import load_workbook

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "palewire"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from New Jersey.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Fire up the cache
    cache = Cache(cache_dir)

    # Download URL
    url = "https://www.nj.gov/labor/assets/PDFs/WARN/WARN_Notice_Archive.xlsx"
    wb_path = cache.download("nj/source.xlsx", url)

    # Read in the workbook
    output_rows = []
    wb = load_workbook(filename=wb_path)
    for ws in wb.worksheets:
        logger.debug(f"Parsing {ws}")
        for i, row in enumerate(ws.rows):

            # Skip header
            if i == 0:
                continue

            # A list with only empty cells will throw an error
            try:
                next(v for v in row if v.value)
            except StopIteration:
                continue

            # Parse out data
            d = {
                "Company": _parse_value(row[0]),
                "City": _parse_value(row[1]),
                "Month Posted": _parse_value(row[2]),
                "Effective Date": _parse_value(row[3]),
                "Workforce Affected": _parse_value(row[4]),
            }

            # Tack it on
            output_rows.append(d)

    # Set the export path
    data_path = data_dir / "nj.csv"

    # Write out the file
    headers = output_rows[0].keys()
    utils.write_dict_rows_to_csv(data_path, headers, output_rows)

    # Return the path to the file
    return data_path


def _parse_value(cell):
    v = cell.value
    if isinstance(v, str):
        return v.strip()
    return v


if __name__ == "__main__":
    scrape()



import logging
import re
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "shallotly"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Maryland.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Set the cache
    cache = Cache(cache_dir)

    # Get the page
    url = "http://www.dllr.state.md.us/employment/warn.shtml"
    r = utils.get_url(url)
    r.encoding = "utf-8"
    html = r.text

    # Save it to the cache
    cache.write("md/source.html", html)

    # Parse the list of links
    soup = BeautifulSoup(html, "html.parser")
    a_list = soup.find_all("a", {"class": "sub"})
    href_list = [a["href"] for a in a_list]

    # Download them all
    html_list = []
    for href in href_list:

        # Request the HTML
        url = f"http://www.dllr.state.md.us/employment/{href}"
        r = utils.get_url(url)
        r.encoding = "utf-8"
        html = r.text

        # Save it to the cache
        cache.write(f"md/{href}.html", html)

        # Add it to the list
        html_list.append(html)

    # Parse them all
    output_rows = []
    for i, html in enumerate(html_list):
        soup = BeautifulSoup(html, "html5lib")

        # Pull out the table
        table_list = soup.find_all("table")
        assert len(table_list) > 0
        table = table_list[0]

        # Get all rows
        row_list = table.find_all("tr")

        # If it's not the first page, slice off the header
        if i > 0:
            row_list = row_list[1:]

        # Loop through all the rows
        for row in row_list:
            # Get the cells
            cell_list = row.find_all("td")

            # Clean them up
            cell_list = [_clean_text(c.text) for c in cell_list]

            # Pass them out
            output_rows.append(cell_list)

    # Set the export path
    data_path = data_dir / "md.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the file
    return data_path


def _clean_text(text):
    """Clean up the provided HTML snippet."""
    if text is None:
        return ""
    text = re.sub(r"\n", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text.strip()


if __name__ == "__main__":
    scrape()



import logging
from pathlib import Path

from openpyxl import load_workbook

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "ydoc5212"]
__tags__ = ["historical", "excel"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Oregon.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Request the page and save it to the cache
    url = (
        "https://storage.googleapis.com/bln-data-public/warn-layoffs/or_historical.xlsx"
    )
    cache = Cache(cache_dir)
    excel_path = cache.download("or/source.xlsx", url)

    # Open it up
    workbook = load_workbook(filename=excel_path)

    # Get the first sheet
    worksheet = workbook.worksheets[0]

    # Convert the sheet to a list of lists
    row_list = []
    # Skip the first two rows, which contain a crufty header
    for r in list(worksheet.rows)[2:]:
        column = [cell.value for cell in r]
        row_list.append(column)

    # Set the export path
    data_path = data_dir / "or.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the file
    return data_path


if __name__ == "__main__":
    scrape()



import logging
import re
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Washington.

    Arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Set the cache
    cache = Cache(cache_dir)

    output_rows = []

    with requests.Session() as session:
        # Request the initial page
        url = "https://fortress.wa.gov/esd/file/warn/Public/SearchWARN.aspx"
        user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:68.0) Gecko/20100101 Firefox/68.0"
        r = utils.get_url(url, user_agent=user_agent, session=session)

        # Save it to the cache
        html = r.text
        cache.write("wa/source.html", html)

        # Parse out the headers
        soup = BeautifulSoup(html, "html5lib")
        table_list = soup.find_all("table")
        first_table = table_list[0]
        first_row = first_table.find_all("tr")[2]
        th_list = first_row.find_all("th")
        headers = [_clean_text(th.text) for th in th_list]
        output_rows.append(headers)

        # Parse the data
        row_list = _parse_table(first_table)
        output_rows.extend(row_list)

        # Start jumping through the pages
        soup_content = BeautifulSoup(r.content, "html5lib")

        page = 2
        while True:
            try:
                # Post for the next page
                formdata = {
                    "__EVENTTARGET": "ucPSW$gvMain",
                    "__EVENTARGUMENT": f"Page${page}",
                    "__VIEWSTATE": soup_content.find(
                        "input", attrs={"name": "__VIEWSTATE"}
                    )["value"],
                    "__EVENTVALIDATION": soup_content.find(
                        "input", attrs={"name": "__EVENTVALIDATION"}
                    )["value"],
                }
                next = session.post(url, data=formdata)
                logger.debug(f"Page status is {next.status_code} for {url}")

                # Update the input variables
                soup_content = BeautifulSoup(next.content, "html5lib")

                # Cache the html
                html = next.text
                cache.write(f"wa/{page}.html", html)

                # Parse out the data
                soup = BeautifulSoup(html, "html5lib")
                table_list = soup.find_all("table")
                first_table = table_list[0]
                row_list = _parse_table(first_table)
                output_rows.extend(row_list)

                # Up the page number
                page += 1

            # Once it fails, we're done
            except Exception:
                break

    # Set the export path
    data_path = data_dir / "wa.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the file
    return data_path


def _parse_table(table) -> list:
    # Parse the cells
    row_list = []
    for row in table.find_all("tr"):
        cell_list = row.find_all(["td"])
        if not cell_list:
            continue
        cell_list = [_clean_text(c.text) for c in cell_list]
        row_list.append(cell_list)

    # Return it with a slice to cut the cruft
    return row_list[2 : len(row_list) - 2]


def _clean_text(text):
    """Clean up the provided HTML snippet."""
    if text is None:
        return ""
    text = re.sub(r"\n", " ", text)
    text = re.sub(r"\s+", " ", text)
    return text.strip()


if __name__ == "__main__":
    scrape()



from pathlib import Path

import pdfplumber
from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["anikasikka"]
__tags__ = ["html", "pdf"]


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Tennessee.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Initialize the cache
    cache = Cache(cache_dir)

    # Grab the HTML page with the latest years of data
    page = utils.get_url(
        "https://www.tn.gov/workforce/general-resources/major-publications0/major-publications-redirect/reports.html"
    )
    html = page.text
    cache.write("tn/source.html", html)

    # Grab the PDF with the archived historial data
    pdf_url = "https://www.tn.gov/content/dam/tn/workforce/documents/majorpublications/reports/WarnReportByMonth.pdf"
    pdf_file = cache.download("tn/pdffile.pdf", pdf_url)

    # Set the headers we'll use for both sources
    tn_headers = [
        "Notice Date",
        "Effective Date",
        "Received Date",
        "Company",
        "City",
        "County",
        "No. Of Employees",
        "Layoff/Closure",
        "Notice ID",
    ]
    cleaned_data = [tn_headers]

    # Parse the latest HTML file and convert to a list of rows, with a header in the first row.
    soup = BeautifulSoup(html, "html5lib")

    # Grab all the list items on the page
    data_list = soup.find_all("p")

    # Loop through them all, skipping the first item, which is a header
    for data in data_list[1:]:

        # splitting the data on its delimiter
        items = str(data).split("|")

        # making sure that the last item in the list is the data value of interest
        # splitting based on last character of each text-html data sequence
        raw_data = []
        for item in items:
            value_html = item.split(":")[-1]
            value_soup = BeautifulSoup(value_html, "html5lib")
            string_list = list(value_soup.stripped_strings)
            value = string_list[-1]
            raw_data.append(value)

        # If there aren't six entries it's junk
        if len(raw_data) != 6:
            continue

        # Pluck out the values we want
        nice_data = [
            raw_data[0],  # Notice Date
            raw_data[4],  # Effective Date
            "",  # Received Date
            raw_data[1],  # Company
            "",  # City
            raw_data[2],  # County
            raw_data[3],  # Number of employees
            "",  # Layoff/Closure
            raw_data[5],  # Notice ID
        ]

        # Add them to the master list
        cleaned_data.append(nice_data)

    # The PDF header blacklist of rows to toss
    pdf_header_blacklist = [
        "Notice Date",
        "Total",
    ]

    # Open the PDF
    with pdfplumber.open(pdf_file) as pdf:

        # Loop through all the pages
        for i, my_page in enumerate(pdf.pages):

            # Sll even pages have data, odd pages don't have the data
            if i % 2 != 0:
                continue

            # Pull out the table and loop through the rows
            table = my_page.extract_table()

            # Cut empty rows
            row_list = [r for r in table if any(r)]

            # If this is a summary table, skip it
            if row_list[0][0].lower().strip() == "summary by month":
                continue

            # Loop through all the rows ...
            for row in row_list:

                # Skip remove redundant headers
                if row[0] in pdf_header_blacklist:
                    continue

                # Toss in an empty Notice ID since it isn't in the PDF
                row.append("")

                # Add the data to our output
                cleaned_data.append(row)

    # Set the path to the final CSV
    output_csv = data_dir / "tn.csv"

    # Write out the rows to the export directory
    utils.write_rows_to_csv(output_csv, cleaned_data)

    # Return the path to the final CSV
    return output_csv


if __name__ == "__main__":
    scrape()



from pathlib import Path

from warn.platforms.job_center.utils import scrape_state

from .. import utils

__authors__ = ["zstumgoren", "Dilcia19"]
__tags__ = ["jobcenter"]


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
    use_cache: bool = True,
) -> Path:
    """
    Scrape data from Kansas.

    Arguments:
    output_dir -- the Path were the result will be saved

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)
    use_cache -- a Boolean indicating whether the cache should be used (default True)

    Returns: the Path where the file is written
    """
    output_csv = data_dir / "ks.csv"
    search_url = "https://www.kansasworks.com/search/warn_lookups"
    # Date chosen based on manual research
    stop_year = 1998
    # Use cache for years before current and prior year
    scrape_state(
        "KS", search_url, output_csv, stop_year, cache_dir, use_cache=use_cache
    )
    return output_csv


if __name__ == "__main__":
    scrape()



import logging
import re
from collections import OrderedDict
from datetime import datetime as dt

from ... import utils
from .site import Site as JobCenterSite

logger = logging.getLogger(__name__)


def scrape_state(
    state_postal, search_url, output_csv, stop_year, cache_dir, use_cache=True
):
    """Date-based scraper for Job Center states.

    This is the primary interface that should be used by downstream scrapers.
    It applies a date-based scraping strategy that:

      - Scrapes one year at a time, in reverse chronological order
      - Always does a fresh scrape for current and prior year
      - Uses cached files for years before current & prior
      - Deduplicates search results

    Args:
        state_postal (str): Two-letter all-caps state postal (e.g. KS)
        search_url (str): Base search url (e.g. https://www.kansasworks.com/search/warn_lookups)
        output_csv (str): Full path to CSV where data should be saved (e.g. ~/.warn-scraper/exports/ks.csv)
        stop_year (int): First year that data is available for state (requires manaul research)
        cache_dir (str): The root directory for WARN's cache files (e.g. ~/.warn-scraper/cache)
        use_cache (boolean, default True): Whether to use cached files for older years

    Returns:
        Full path to exported csv (e.g. ~/.warn-scraper/exports/ks.csv)
    """
    yearly_dates = _date_ranges_to_scrape(stop_year)

    # No caching should be used for current and prior year, so
    # we have to separate those from remaining years.
    no_cache_years = [yearly_dates.pop(0), yearly_dates.pop(0)]

    # Set up scraper instance
    state_cache_dir = cache_dir / state_postal.lower()
    site = JobCenterSite(state_postal.upper(), search_url, cache_dir=state_cache_dir)

    # Date-based searches produce search result pages that appear to have certain
    # records duplicated over paged results. We'll initially write all data to a raw
    # file which we then deduplicate to produce the final output_csv.
    raw_csv = cache_dir / f"{state_postal.lower()}_raw.csv"
    logger.debug(f"Generating {raw_csv}")

    # 0. Write header row first
    headers = [
        "employer",
        "notice_date",
        "number_of_employees_affected",
        "warn_type",
        "city",
        "zip",
        "lwib_area",
        "address",
        "record_number",
        "detail_page_url",
    ]
    utils.write_rows_to_csv(raw_csv, [headers])
    # Execute the scrape in two batches
    # 1. Current and prior year. Always scrape fresh (i.e. never use cached files)
    #    in case records have been updated.
    _scrape_years(site, raw_csv, headers, no_cache_years, use_cache=False)
    # 2. Years before current & prior, going back to stop_year.
    #    We should generally use cached files for these older years,
    #    since data is less likely to be updated.
    _scrape_years(site, raw_csv, headers, yearly_dates, use_cache=use_cache)
    _dedupe(raw_csv, output_csv)
    return output_csv


def _scrape_years(site, output_csv, headers, start_end_dates, use_cache=True):
    """Loop through years of data and write out to CSV."""
    # NOTE: Scraping for Jan 1 - Dec 31 for current year works
    # throughout the year. Additionally, it allows us to avoid
    # generating cache files for all days of the year.
    for start, end in start_end_dates:
        kwargs = {
            "start_date": start,
            "end_date": end,
            "use_cache": use_cache,
        }
        pages_dict, data = site.scrape(**kwargs)
        rows = [_prepare_row(row) for row in data]
        # We previously wrote the header so use append mode for data rows
        utils.write_dict_rows_to_csv(output_csv, headers, rows, mode="a")


def _prepare_row(row):
    """Flatten the nested dict for downstream export to CSV."""
    # Returned data includes fields from search result page and
    # data from detail page record for each layoff notice
    # the latter contains two key fields (address and number affected)
    detail = row.pop("detail")
    row["number_of_employees_affected"] = detail["number_of_employees_affected"]
    row["address"] = re.sub(r"\n+", "; ", detail["address"].strip())
    row["record_number"] = detail["record_number"]
    return row


def _date_ranges_to_scrape(stop_year):
    """Generate a list of start/end pairs from current year to some year in the past."""
    start = "{}-01-01"
    end = "{}-12-31"
    current_year = dt.today().year
    years = sorted(range(stop_year, current_year + 1), reverse=True)
    yearly_dates = []
    for year in years:
        yearly_dates.append((start.format(year), end.format(year)))
    return yearly_dates


def _dedupe(raw_csv, output_csv):
    """Create an ordered dict to discard dupes while preserving row order."""
    data = OrderedDict()
    raw_count = 0
    utils.create_directory(raw_csv, is_file=True)
    with open(raw_csv, newline="") as src:
        for row in src:
            raw_count += 1
            data[row] = row
    # Write the deduped rows to final output_csv
    final_count = 0
    utils.create_directory(output_csv, is_file=True)
    with open(output_csv, "w", newline="") as out:
        for row in data.keys():
            final_count += 1
            out.write(row)
    num_removed = raw_count - final_count
    if num_removed > 0:
        logger.debug(f"Removed {num_removed} duplicate records from {raw_csv}")
    return output_csv



import logging
from pathlib import Path

from bs4 import BeautifulSoup
from openpyxl import load_workbook

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "ydoc5212", "chriszs"]
__tags__ = ["excel"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Rhode Island.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Open the cache
    cache = Cache(cache_dir)

    state_code = "ri"

    # Get the HTML
    base_url = "https://dlt.ri.gov/"
    url = f"{base_url}/employers/worker-adjustment-and-retraining-notification-warn"
    r = utils.get_url(url)
    html = r.text
    cache.write(f"{state_code}/source.html", html)

    # Find links
    soup = BeautifulSoup(html, "html.parser")
    links = soup.find_all("a")

    row_list = []

    for link in links:
        if "WARN Report" in link.text:
            excel_url = f"{base_url}{link.get('href')}"
            excel_path = cache.download(f"{state_code}/WARN Report.xlsx", excel_url)

            # Open it up
            workbook = load_workbook(filename=excel_path)

            # Get the first sheet
            worksheet = workbook.worksheets[0]

            for r in list(worksheet.rows)[3:]:
                column = [cell.value for cell in r]
                row_list.append(column)

    # Write out
    data_path = data_dir / "ri.csv"
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the CSV
    return data_path


if __name__ == "__main__":
    scrape()



import logging
import re
from datetime import datetime
from pathlib import Path

from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "ydoc5212"]
__tags__ = ["html"]

logger = logging.getLogger(__name__)


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Wisconsin.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Set the cache
    cache = Cache(cache_dir)

    # Get the current year
    today = datetime.today()
    current_year = today.year

    # Set the date range we're going to scrape
    year_range = list(range(2016, current_year + 1))
    year_range.reverse()

    # Loop through the years and download the pages
    html_list = []
    for year in year_range:
        # Since the 2022 page doesn't exist yet, we're going to hack in a skip
        if year == 2022:
            continue

        # Request fresh pages, use cache for old ones
        cache_key = f"wi/{year}.html"
        if cache.exists(cache_key) and year < current_year - 1:
            html = cache.read(cache_key)
        else:
            url = f"https://dwd.wisconsin.gov/dislocatedworker/warn/{year}/default.htm"
            r = utils.get_url(url)
            html = r.text
            cache.write(cache_key, html)

        # Add to the list
        html_list.append(html)

    output_rows = []
    for ihtml, html in enumerate(html_list):
        # Parse the HTML
        soup = BeautifulSoup(html, "html5lib")

        # Fish out the tables
        table_list = soup.find_all("table")

        # Remove the "Updates to Previously Filed Notices" tables
        # We can single them out because they only have two columns
        notice_tables = [t for t in table_list if len(t.find("tr").find_all("th")) > 2]

        # Loop through the tables
        for itable, table in enumerate(notice_tables):
            # Get all the rows
            row_list = table.find_all("tr")

            # If this is the first table on the first page, get headers too
            tags = ["td"]
            if ihtml == 0 and itable == 0:
                tags.append("th")

            for row in row_list:
                # Pull out the cells and clean them
                cell_list = [_clean_text(c.text) for c in row.find_all(tags)]

                # Skip empty rows
                try:
                    # A list with only empty cell will throw an error
                    next(i for i in cell_list if i)
                except StopIteration:
                    continue

                # Tack what we've got onto the pile
                output_rows.append(cell_list)

    # Set the export path
    data_path = data_dir / "wi.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, output_rows)

    # Return the path to the file
    return data_path


def _clean_text(text):
    """Clean up the provided cell."""
    # remove trailing characters after LayoffBeginDate
    if re.match(r"^[0-9]{1,2}/[0-9]{1,2}/[0-9]{4}", text):
        text = re.sub(r"(?<=[0-9]{4}).*", "", text)
    return text.strip()


if __name__ == "__main__":
    scrape()



#!/usr/bin/env python
"""Configure the package for distribution."""
import distutils.cmd
import os
from importlib import import_module
from pathlib import Path

import jinja2
import us
from setuptools import find_packages, setup

import warn


def read(file_name):
    """Read the provided file."""
    this_dir = os.path.dirname(__file__)
    file_path = os.path.join(this_dir, file_name)
    with open(file_path) as f:
        return f.read()


def version_scheme(version):
    """
    Version scheme hack for setuptools_scm.

    Appears to be necessary to due to the bug documented here: https://github.com/pypa/setuptools_scm/issues/342

    If that issue is resolved, this method can be removed.
    """
    import time

    from setuptools_scm.version import guess_next_version

    if version.exact:
        return version.format_with("{tag}")
    else:
        _super_value = version.format_next_version(guess_next_version)
        now = int(time.time())
        return _super_value + str(now)


def local_version(version):
    """
    Local version scheme hack for setuptools_scm.

    Appears to be necessary to due to the bug documented here: https://github.com/pypa/setuptools_scm/issues/342

    If that issue is resolved, this method can be removed.
    """
    return ""


class TallyCommand(distutils.cmd.Command):
    """Tally scrapers and update sources dashboard in the docs."""

    description = "Tally scrapers and update sources dashboard in the docs"
    user_options: list = []

    def initialize_options(self):
        """Set default values for options."""
        pass

    def finalize_options(self):
        """Finalize values."""
        pass

    def run(self):
        """Run command."""
        print("Tallying sources")

        this_dir = Path(__file__).parent
        loader = jinja2.FileSystemLoader(searchpath=this_dir / "docs" / "_templates/")
        env = jinja2.Environment(loader=loader)

        scraper_list = warn.utils.get_all_scrapers()
        print(f"{len(scraper_list)} scrapers found")

        docs_dir = this_dir / "docs" / "scrapers"
        has_docs = [f.stem for f in docs_dir.glob("*.md") if f.stem in scraper_list]
        print(f"{len(has_docs)} scrapers have docs")

        haves, have_nots = [], []
        target_list = us.states.STATES_AND_TERRITORIES + [us.states.DC]
        for t in target_list:
            state = t.__dict__
            abbr = state["abbr"].lower()
            if abbr in scraper_list:
                state["has_docs"] = abbr in has_docs
                module = import_module(f"warn.scrapers.{abbr}")
                state["authors"] = sorted(module.__authors__)
                state["tags"] = sorted(module.__tags__)
                haves.append(state)
            else:
                state["has_docs"] = False
                state["authors"] = None
                state["tags"] = None
                have_nots.append(state)
        print(f"{len(haves)} states and territories have a scraper")
        print(f"{len(have_nots)} states and territories do not have a scraper")

        context = {
            "haves": haves,
            "have_nots": have_nots,
            "targets": target_list,
        }
        template = env.get_template("sources.md.tmpl")
        md = template.render(**context)

        with open(this_dir / "docs" / "sources.md", "w") as fh:
            fh.write(md)


setup(
    name="warn-scraper",
    description="Command-line interface for downloading WARN Act notices of qualified plant closings and mass layoffs from state government websites",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Big Local News",
    url="https://github.com/biglocalnews/warn-scraper",
    packages=find_packages(),
    include_package_data=True,
    entry_points="""
        [console_scripts]
        warn-scraper=warn.cli:main
    """,
    install_requires=[
        "click",
        "bs4",
        "html5lib",
        "pdfplumber",
        "requests",
        "openpyxl",
        "xlrd",
        "tenacity",
    ],
    license="Apache 2.0 license",
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
    tests_require=[
        "pytest",
        "pytest-vcr",
    ],
    setup_requires=["pytest-runner", "setuptools_scm"],
    use_scm_version={"version_scheme": version_scheme, "local_scheme": local_version},
    project_urls={
        "Documentation": "https://warn-scraper.readthedocs.io",
        "Maintainer": "https://github.com/biglocalnews",
        "Source": "https://github.com/biglocalnews/warn-scraper",
        "Tracker": "https://github.com/biglocalnews/warn-scraper/issues",
    },
    cmdclass={
        "tallysources": TallyCommand,
    },
)



import datetime
import logging
import re
from os.path import exists
from pathlib import Path

import pdfplumber
import requests
import tenacity
import urllib3
from bs4 import BeautifulSoup

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "Dilcia19", "shallotly"]
__tags__ = ["html", "pdf"]

logger = logging.getLogger(__name__)


FIELDS = [
    "Company Name",
    "State Notification Date",
    "Layoff Date",
    "Employees Affected",
    "Industry",
    "Attachment",
]
CSV_HEADERS = FIELDS[:-1]  # Clip the Attachment header


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Florida.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    output_csv = data_dir / "fl.csv"
    cache = Cache(cache_dir)  # ~/.warn-scraper/cache
    # FL site requires realistic User-Agent.
    headers = {
        "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36"
    }
    url = "http://floridajobs.org/office-directory/division-of-workforce-services/workforce-programs/reemployment-and-emergency-assistance-coordination-team-react/warn-notices"
    response = requests.get(url, headers=headers, verify=False)
    logger.debug(f"Request status is {response.status_code} for {url}")
    soup = BeautifulSoup(response.text, "html.parser")
    # find & visit each year's WARN page
    base_url = "http://reactwarn.floridajobs.org/WarnList/"
    links = soup.find_all("a", href=re.compile(base_url))
    href_lookup = {_extract_year(link.text): link.get("href") for link in links}

    # Loop through years and add any missing to the lookup
    most_recent_year = int(list(href_lookup.keys())[0])
    earliest_year = 2015  # We expect files to be available for at least 2015
    for year in range(earliest_year, most_recent_year):
        if str(year) not in href_lookup:
            href_lookup[str(year)] = f"{base_url}viewPreviousYearsPDF?year={year}"

    output_rows = []
    # Loop through years and scrape data
    for year_url in href_lookup.values():
        if "PDF" in year_url:
            rows_to_add = _scrape_pdf(cache, cache_dir, year_url, headers)
        else:
            html_pages = _scrape_html(cache, year_url, headers)
            rows_to_add = _html_to_rows(html_pages)
        # Convert rows to dicts
        rows_as_dicts = [dict(zip(FIELDS, row)) for row in rows_to_add]
        output_rows.extend(rows_as_dicts)
    utils.write_dict_rows_to_csv(
        output_csv, CSV_HEADERS, output_rows, extrasaction="ignore"
    )
    return output_csv


# scrapes each html page for the current year
# returns a list of the year's html pages
# note: no max amount of retries (recursive scraping)
@tenacity.retry(
    wait=tenacity.wait_exponential(),
    retry=tenacity.retry_if_exception_type(requests.HTTPError),
)
def _scrape_html(cache, url, headers, page=1):
    urllib3.disable_warnings()  # sidestep SSL error
    # extract year from URL
    year = _extract_year(url)
    html_cache_key = f"fl/{year}_page_{page}.html"
    current_year = datetime.date.today().year
    last_year = str(current_year - 1)
    current_year = str(current_year)
    page_text = ""
    # search in cache first before scraping
    try:
        # re-scrape current year by default
        if not (year == current_year or year == last_year):
            logger.debug(f"Trying to read from cache: {html_cache_key}")
            cachefile = cache.read(html_cache_key)
            page_text = cachefile
            logger.debug(f"Page fetched from cache: {html_cache_key}")
        else:
            raise FileNotFoundError
    except FileNotFoundError:
        # scrape & cache html
        response = requests.get(url, headers=headers, verify=False)
        logger.debug(f"Request status is {response.status_code} for {url}")
        response.raise_for_status()
        page_text = response.text
        cache.write(html_cache_key, page_text)
        logger.debug(f"Successfully scraped page {url} to cache: {html_cache_key}")
    page_text = page_text.replace("</br>", "\n")
    # search the page we just scraped for links to the next page
    soup = BeautifulSoup(page_text, "html.parser")
    footer = soup.find("tfoot")
    if footer:
        next_page = page + 1
        nextPageLink = footer.find(
            "a", href=re.compile(f"page={next_page}")
        )  # find link to next page, if exists
        # recursively scrape until we have a list of all the pages' html
        if nextPageLink:
            url = "http://reactwarn.floridajobs.org" + nextPageLink.get(
                "href"
            )  # /WarnList/Records?year=XXXX&page=X
            # recursively make list of all the next pages' html
            pages_html = _scrape_html(cache, url, headers, next_page)
            # add the current page to the recursive list
            pages_html.append(page_text)
            return pages_html
    # last page reached
    return [page_text]


def _html_to_rows(page_text):
    """Extract data rows from list of html pages."""
    output_rows = []
    for page in page_text:
        soup = BeautifulSoup(page, "html5lib")
        table = soup.find("table")
        # extract table data
        tbody = table.find("tbody")
        for table_row in tbody.find_all("tr"):
            columns = table_row.find_all("td")
            output_row = []
            for column in columns:
                output_row.append(column.text.strip())
            output_rows.append(output_row)
    return output_rows


# download and scrape pdf
@tenacity.retry(
    wait=tenacity.wait_exponential(),
    retry=tenacity.retry_if_exception_type(requests.HTTPError),
)
def _scrape_pdf(cache, cache_dir, url, headers):
    # sidestep SSL error
    urllib3.disable_warnings()
    # extract year from URL
    year = _extract_year(url)
    pdf_cache_key = f"fl/{year}.pdf"
    download = ""
    # download pdf if not in the cache
    if not exists(pdf_cache_key):
        response = requests.get(url, headers=headers, verify=False)
        logger.debug(f"Request status is {response.status_code} for {url}")
        response.raise_for_status()
        # download & cache pdf
        download = response.content
        with open(f"{cache_dir}/{pdf_cache_key}", "wb") as f:
            f.write(download)
        logger.debug(f"Successfully scraped PDF from {url} to cache: {pdf_cache_key}")
    # scrape tables from PDF
    with pdfplumber.open(f"{cache_dir}/{pdf_cache_key}") as pdf:
        pages = pdf.pages
        output_rows = []
        for page_num, page in enumerate(pages):
            table = page.extract_table(table_settings={})
            # remove each year's header
            if page_num == 0:
                table.pop(0)
            table = _clean_table(table, output_rows)
            output_rows.extend(table)  # merging lists
    logger.debug(f"Successfully scraped PDF from {url}")
    return output_rows


# adds split rows to output_rows by reference, returns list of page's rows to be added
def _clean_table(table, all_rows):
    table_rows = []
    for row_idx, row in enumerate(table):
        current_row = []
        # fix row splitting b/n pages sometimes
        if _is_multiline_row(row_idx, row):
            if (
                len(row) > 5
            ):  # fix where both row is split AND columns skewed right (like page 14 of 2016.pdf)
                row = [row[0], row[1], row[2], row[3], row[6]]
            for field_idx, field_to_add in enumerate(row):
                if field_to_add:
                    all_rows[-1][
                        field_idx
                    ] += field_to_add  # MERGE fields with last row from all_rows (i.e. the last row from prior page)
            continue
        for field_idx, field in enumerate(row):
            # ignore any redundant header rows
            if _is_header_row(field_idx, field):
                break
            # fix column skew by skipping blank columns
            if field:
                clean_field = field.strip()
                if clean_field:
                    current_row.append(clean_field)
        if current_row:
            table_rows.append(current_row)
    return table_rows


def _is_multiline_row(row_idx, row):
    # this is a row that has been split between pages
    # we want to remedy this split in the output data
    return row_idx == 0 and row[1] == "" and row[3] == ""


def _is_header_row(field_idx, field):
    # we already have a header management strategy
    # but there are still erroneous redundant headers strewn about
    # and we need to remove them from the data
    # because we only need 1 header row.
    return field_idx == 0 and field == "COMPANY NAME"


def _extract_year(text):
    """Extract the year from the string."""
    if text is None:
        return None
    return re.search(r"(\d{4})", text).group(1)


if __name__ == "__main__":
    scrape()



from pathlib import Path

from bs4 import BeautifulSoup
from openpyxl import load_workbook

from .. import utils
from ..cache import Cache

__authors__ = ["zstumgoren", "ydoc5212"]
__tags__ = ["html", "excel"]


def scrape(
    data_dir: Path = utils.WARN_DATA_DIR,
    cache_dir: Path = utils.WARN_CACHE_DIR,
) -> Path:
    """
    Scrape data from Montana.

    Keyword arguments:
    data_dir -- the Path were the result will be saved (default WARN_DATA_DIR)
    cache_dir -- the Path where results can be cached (default WARN_CACHE_DIR)

    Returns: the Path where the file is written
    """
    # Get the URL
    url = "https://wsd.dli.mt.gov/wioa/related-links/warn-notice-page"
    r = utils.get_url(url)
    html = r.text

    # Save it to the cache
    cache = Cache(cache_dir)
    cache.write("mt/source.html", html)

    # Parse out the Excel link
    soup = BeautifulSoup(html, "html.parser")
    links = soup.find(id="boardPage").find_all("a")
    excel_name = [
        link.attrs["href"]
        for link in links
        if link.attrs.get("href", "").endswith(
            "xlsx"
        )  # URL will look like: ="../../_docs/wioa/warn-9-1-21.xlsx"
    ][0].split("/")[-1]
    excel_url = f"https://wsd.dli.mt.gov/_docs/wioa/{excel_name}"

    # Download the Excel file
    excel_path = cache.download("mt/source.xlsx", excel_url)

    # Open it up
    workbook = load_workbook(filename=excel_path)

    # Get the first sheet
    worksheet = workbook.worksheets[0]

    # Convert the sheet to a list of lists
    row_list = []
    for row in worksheet.rows:
        # Get the cells
        cell_list = [cell.value for cell in row]

        # Skip empty rows
        try:
            # A list with only empty cell will throw an error
            next(i for i in cell_list if i)
        except StopIteration:
            continue

        # Add it to the pile
        row_list.append(cell_list)

    # Set the export path
    data_path = data_dir / "mt.csv"

    # Write out the file
    utils.write_rows_to_csv(data_path, row_list)

    # Return the path to the file
    return data_path


if __name__ == "__main__":
    scrape()
